from __future__ import annotations

from pathlib import Path
from typing import Union

from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.model import Model, ModelfitResultsError
from pharmpy.modeling import (
    get_ids,
    get_model_covariates,
    get_number_of_individuals,
    get_number_of_observations,
    get_number_of_observations_per_individual,
)
from pharmpy.workflows.results import ModelfitResults


def _all_parameters(model: Model, res: ModelfitResults):
    # All parameter estimates including fixed
    # This might warrant a function in modeling
    d = {p.name: p.init for p in model.parameters}
    parameter_estimates = res.parameter_estimates
    assert parameter_estimates is not None
    d.update(dict(parameter_estimates))
    return pd.Series(d)


def _create_dataset(model: Model, res: ModelfitResults):
    if res is None:
        raise ModelfitResultsError("Need ModelfitResults for model")
    if res.residuals is None:
        raise ModelfitResultsError("No residuals available in ModelfitResults")
    if 'CWRES' not in res.residuals.columns:
        raise ModelfitResultsError("CWRES not an available residual in ModelfitResults")
    idcol = model.datainfo.id_column.name
    nids = get_number_of_individuals(model)
    nobs = get_number_of_observations(model)
    nobsi = get_number_of_observations_per_individual(model)
    cwres = res.residuals['CWRES']
    iofv = res.individual_estimates
    npar = len(model.parameters)
    ofv = res.ofv
    assert ofv is not None

    # Max ratio of abs(ETAi) and omegai
    variance_omegas = [
        model.random_variables[rv].get_variance(rv).name for rv in model.random_variables.etas.names
    ]
    all_paramvals = _all_parameters(model, res)
    omega_estimates = np.sqrt(all_paramvals[variance_omegas])
    individual_estimates = res.individual_estimates
    assert individual_estimates is not None
    abs_ebes = individual_estimates.abs()
    if omega_estimates.empty:  # No etas in model
        max_ebe_ratio = 1.0
    else:
        ebe_ratio = abs_ebes / list(omega_estimates)
        max_ebe_ratio = ebe_ratio.max(axis=1).fillna(1.0)
        # Set to 1 if division by zero, e.g. from omega fix to 0

    # exp(OFVi / nobsi) / exp(OFV / nobs)
    iofv = res.individual_ofv
    ofv_ratio = np.exp(iofv / nobsi) / np.exp(ofv / nobs)

    # mean(ETA / OMEGA)
    cov = res.individual_estimates_covariance
    assert cov is not None
    etc_diag = np.sqrt(pd.DataFrame([np.diag(y) for y in cov], columns=cov.iloc[0].columns))
    if omega_estimates.empty:
        mean_etc_ratio = 1.0
    else:
        etc_ratio = etc_diag / list(omega_estimates)
        mean_etc_ratio = etc_ratio.mean(axis=1).fillna(1.0)
        mean_etc_ratio.index = ofv_ratio.index

    # max((abs(indcov - mean(cov))) / sd(cov))
    cov_names = get_model_covariates(model, strings=True)
    if len(cov_names) > 0:
        covariates = model.dataset[cov_names + [idcol]].set_index(idcol)
        mean_covs = covariates.groupby(idcol).mean()
        maxcov = (abs(mean_covs - mean_covs.mean()) / mean_covs.std()).max(axis=1)
    else:
        maxcov = 0.0

    df = pd.DataFrame(
        {
            'nids': nids,
            'nobs': nobs,
            'nobs_subj': nobsi / (nobs / nids),
            'nobsi': nobsi,
            'ncovs': len(cov_names),
            'maxcov': maxcov,
            'max_cwres': abs(cwres).groupby(idcol).max(),
            'median_cwres': abs(cwres).groupby(idcol).median(),
            'max_ebe_ratio': max_ebe_ratio,
            'ofv_ratio': ofv_ratio,
            'etc_ratio': mean_etc_ratio,
            'iofv': iofv,
            'npar': npar,
            'ofv': ofv,
        }
    )
    return df


def predict_outliers(model: Model, results: ModelfitResults, cutoff: float = 3.0):
    """Predict outliers for a model using a machine learning model.

    See the :ref:`simeval <Individual OFV summary>` documentation for a definition of the `residual`

    Parameters
    ----------
    model : Model
        Pharmpy model
    results : ModelfitResults
        ModelfitResults for the model
    cutoff : float
        Cutoff threshold for a residual singalling an outlier

    Returns
    -------
    pd.DataFrame
        Dataframe over the individuals with a `residual` column containing the raw predicted
        residuals and a `outlier` column with a boolean to tell whether the individual is
        an outlier or not.

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> results = model.modelfit_results
    >>> predict_outliers(model, results)     # doctest: +SKIP
        residual  outlier
    ID
    1  -0.281443    False
    2   1.057392    False
    3  -0.119105    False
    4  -0.846849    False
    5   0.600540    False
    6   1.014008    False
    7  -0.750734    False
    8   0.247175    False
    9   0.117002    False
    10 -0.835389    False
    11  3.529582     True
    12 -0.035670    False
    13  0.292333    False
    14  0.303278    False
    15 -0.565949    False
    16 -0.078192    False
    17 -0.291295    False
    18  2.466421    False
    19 -0.402343    False
    20 -0.699996    False
    21 -0.567987    False
    22 -0.526776    False
    23  0.303918    False
    24 -0.177588    False
    25  1.272142    False
    26 -0.390000    False
    27  0.775876    False
    28 -0.501528    False
    29 -0.700951    False
    30 -0.352599    False
    31  0.294196    False
    32  0.744014    False
    33 -0.215364    False
    34  0.208691    False
    35  1.713130    False
    36  0.300293    False
    37 -0.810736    False
    38  0.459877    False
    39 -0.675125    False
    40 -0.563835    False
    41 -0.526945    False
    42  4.449199     True
    43  0.720714    False
    44 -0.792248    False
    45 -0.860923    False
    46 -0.731858    False
    47 -0.247131    False
    48  1.894190    False
    49 -0.282737    False
    50 -0.153398    False
    51  1.200546    False
    52  0.902774    False
    53  0.586427    False
    54  0.183329    False
    55  1.036930    False
    56 -0.639868    False
    57 -0.765279    False
    58 -0.209665    False
    59 -0.225693    False

    See also
    --------
    predict_influential_individuals
    predict_influential_outliers

    """
    model_path = Path(__file__).resolve().parent / 'ml_models' / 'outliers.tflite'
    data = _create_dataset(model, results)
    output = _predict_with_tflite(model_path, data)
    df = pd.DataFrame({'residual': output, 'outlier': output > cutoff}, index=get_ids(model))
    df.index.name = model.datainfo.id_column.name
    return df


def predict_influential_individuals(model: Model, results: ModelfitResults, cutoff: float = 3.84):
    """Predict influential individuals for a model using a machine learning model.

    Parameters
    ----------
    model : Model
        Pharmpy model
    results : ModelfitResults
        Results for model
    cutoff : float
        Cutoff threshold for a dofv signalling an influential individual

    Returns
    -------
    pd.DataFrame
        Dataframe over the individuals with a `dofv` column containing the raw predicted
        delta-OFV and an `influential` column with a boolean to tell whether the individual is
        influential or not.

    See also
    --------
    predict_influential_outliers
    predict_outliers

    """

    model_path = Path(__file__).resolve().parent / 'ml_models' / 'infinds.tflite'
    data = _create_dataset(model, results)
    output = _predict_with_tflite(model_path, data)
    df = pd.DataFrame({'dofv': output, 'influential': output > cutoff}, index=get_ids(model))
    df.index.name = model.datainfo.id_column.name
    return df


def predict_influential_outliers(
    model: Model,
    results: ModelfitResults,
    outlier_cutoff: float = 3.0,
    influential_cutoff: float = 3.84,
):
    """Predict influential outliers for a model using a machine learning model.

    Parameters
    ----------
    model : Model
        Pharmpy model
    results : ModelfitResults
        Results for model
    outlier_cutoff : float
        Cutoff threshold for a residual signalling an outlier
    influential_cutoff : float
        Cutoff threshold for a dofv signalling an influential individual

    Returns
    -------
    pd.DataFrame
        Dataframe over the individuals with a `outliers` and `dofv` columns containing the raw
        predictions and `influential`, `outlier` and `influential_outlier` boolean columns.

    See also
    --------
    predict_influential_individuals
    predict_outliers

    """

    outliers = predict_outliers(model, results, cutoff=outlier_cutoff)
    infinds = predict_influential_individuals(model, results, cutoff=influential_cutoff)
    df = pd.concat([outliers, infinds], axis=1)
    df['influential_outlier'] = df['outlier'] & df['influential']
    return df


def _predict_with_tflite(model_path: Union[str, Path], data: pd.DataFrame):
    import tflite_runtime.interpreter as tflite

    interpreter = tflite.Interpreter(str(model_path))
    interpreter.allocate_tensors()
    input_details = interpreter.get_input_details()
    output_details = interpreter.get_output_details()

    npdata = data.astype('float32').values
    nrows = len(data)
    output = np.empty(nrows)

    for i in range(0, nrows):
        interpreter.set_tensor(input_details[0]['index'], npdata[i : (i + 1), :])
        interpreter.invoke()
        output_data = interpreter.get_tensor(output_details[0]['index'])
        output[i] = output_data[0][0]

    return output



# The (legacy) NONMEM FCON Model class
# This module was created to allow reading in FDATA files
# using the format in the FCONS for testing and verification
# purposes.

import re
from dataclasses import dataclass
from pathlib import Path

from pharmpy.deps import pandas as pd
from pharmpy.model import Model as BaseModel
from pharmpy.model import ModelSyntaxError


def detect_model(src, *args, **kwargs):
    """Check if src represents a legacy NONMEM control stream"""
    if isinstance(src, str) and src.startswith("FILE    FSTREAM"):
        return Model
    else:
        return None


@dataclass
class FCONInternals:
    code: str = None
    path: Path = None


class Model(BaseModel):
    def __init__(self, **kwargs):
        self._internals = kwargs['internals']
        self._dataset = kwargs['dataset']


def _parse_labels_and_formats(code):
    in_labl = False
    labels = []
    lines = iter(code.split('\n'))
    for line in lines:
        if line.startswith('LABL'):
            in_labl = True
        if in_labl:
            stripped = line[4:].replace(' ', '')
            if stripped.startswith(','):
                in_labl = False
            else:
                a = stripped.split(',')
                labels.extend(a)
        if line.startswith('FORM'):
            next_line = next(lines)
            formats = next_line.strip()[1:-1].split(',')
            break
    else:
        raise ModelSyntaxError("Problems parsing the FORM record in FCONS")

    return labels, formats


def parse_dataset(code, path):
    labels, formats = _parse_labels_and_formats(code)
    widths = []
    for fmt in formats:
        m = re.match(r'(\d+)\w(\d+)', fmt)
        if not m:
            raise ModelSyntaxError(f"Unrecognized Data format in FORM: {fmt}")
        widths += [int(m.group(2))] * int(m.group(1))
    df = pd.read_fwf(path.parent / 'FDATA', widths=widths, header=None, names=labels)
    return df


def parse_model(code: str, path: Path):
    internals = FCONInternals(code=code, path=path)
    dataset = parse_dataset(code, path)
    return Model(internals=internals, dataset=dataset)



import re
import shutil

import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Assignment
from pharmpy.modeling import (
    has_additive_error_model,
    has_combined_error_model,
    has_proportional_error_model,
    has_weighted_error_model,
    read_model,
    read_model_from_string,
    remove_error_model,
    set_additive_error_model,
    set_combined_error_model,
    set_dtbs_error_model,
    set_iiv_on_ruv,
    set_power_on_ruv,
    set_proportional_error_model,
    set_weighted_error_model,
    use_thetas_for_error_stdev,
)
from pharmpy.modeling.error import _get_prop_init, set_time_varying_error_model


def test_remove_error_model(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = remove_error_model(model)
    assert model.model_code.split('\n')[11] == 'Y = F'


def test_set_additive_error_model(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = set_additive_error_model(model)
    assert model.model_code.split('\n')[11] == 'Y = F + EPS(1)'
    assert model.model_code.split('\n')[17] == '$SIGMA  11.2225 ; sigma'
    before = model.model_code
    model = set_additive_error_model(model)  # One more time and nothing should change
    assert before == model.model_code

    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_dvid.mod')
    model = set_additive_error_model(model)
    rec = model.internals.control_stream.get_records('ERROR')[0]
    correct = """$ERROR
Y_1 = F + EPS(3)
Y_2 = F + EPS(1)*F + EPS(2)

IF (DVID.EQ.1) Y = Y_1
IF (DVID.EQ.2) Y = Y_2
"""
    assert str(rec) == correct


def test_set_additive_error_model_logdv(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = set_additive_error_model(model, data_trans="log(Y)")
    assert model.model_code.split('\n')[11] == 'Y = LOG(F) + EPS(1)/F'
    assert model.model_code.split('\n')[17] == '$SIGMA  11.2225 ; sigma'


def test_set_proportional_error_model_nolog(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = model.replace(
        statements=model.statements[0:5] + Assignment.create('Y', 'F') + model.statements[6:]
    )
    model = set_proportional_error_model(model)
    assert model.model_code.split('\n')[16] == 'Y = F + EPS(1)*IPREDADJ'
    assert model.model_code.split('\n')[22] == '$SIGMA  0.09 ; sigma'

    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = set_proportional_error_model(model, zero_protection=False)
    assert model.model_code.split('\n')[11] == 'Y=F+F*EPS(1)'
    assert model.model_code.split('\n')[17] == '$SIGMA 0.013241'

    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_dvid.mod')
    model = set_proportional_error_model(model, dv=2)
    rec = model.internals.control_stream.get_records('ERROR')[0]
    correct = """$ERROR
Y_1 = F + EPS(1)*F
IF (F.EQ.0) THEN
    IPREDADJ = 2.22500000000000E-16
ELSE
    IPREDADJ = F
END IF
Y_2 = F + EPS(2)*IPREDADJ

IF (DVID.EQ.1) Y = Y_1
IF (DVID.EQ.2) Y = Y_2
"""
    assert str(rec) == correct


def test_set_proportional_error_model_log(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = model.replace(
        statements=model.statements[0:5] + Assignment.create('Y', 'F') + model.statements[6:]
    )
    model = set_proportional_error_model(model, data_trans='log(Y)')
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2

$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V

$ERROR
IF (F.EQ.0) THEN
    IPREDADJ = 2.22500000000000E-16
ELSE
    IPREDADJ = F
END IF
Y = LOG(IPREDADJ) + EPS(1)

$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA  0.09 ; sigma
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_set_combined_error_model(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = set_combined_error_model(model)
    assert model.model_code.split('\n')[11] == 'Y = F + EPS(1)*F + EPS(2)'
    assert model.model_code.split('\n')[17] == '$SIGMA  0.09 ; sigma_prop'
    assert model.model_code.split('\n')[18] == '$SIGMA  11.2225 ; sigma_add'
    before = model.model_code
    model = set_combined_error_model(model)  # One more time and nothing should change
    assert before == model.model_code


def test_set_combined_error_model_log(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = set_combined_error_model(model, data_trans='log(Y)')
    assert model.model_code.split('\n')[11] == 'Y = LOG(F) + EPS(2)/F + EPS(1)'
    assert model.model_code.split('\n')[17] == '$SIGMA  0.09 ; sigma_prop'
    assert model.model_code.split('\n')[18] == '$SIGMA  11.2225 ; sigma_add'


def test_set_combined_error_model_with_eta_on_ruv(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = set_iiv_on_ruv(model)
    model = set_combined_error_model(model)
    assert model.model_code.split('\n')[12] == 'Y = F + EPS(1)*F*EXP(ETA_RV1) + EPS(2)*EXP(ETA_RV1)'


def test_set_combined_error_model_with_time_varying(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = set_time_varying_error_model(model, cutoff=1.0)
    model = set_combined_error_model(model)
    assert model.model_code.split('\n')[11] == 'IF (TIME.LT.1.0) THEN'
    assert model.model_code.split('\n')[12] == '    Y = F + EPS(1)*F*THETA(3) + EPS(2)*THETA(3)'
    assert model.model_code.split('\n')[14] == '    Y = F + EPS(1)*F + EPS(2)'


def test_set_combined_error_model_with_time_varying_and_eta_on_ruv(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = set_time_varying_error_model(model, cutoff=1.0)
    model = set_iiv_on_ruv(model)
    model = set_combined_error_model(model)
    assert (
        model.model_code.split('\n')[13]
        == '    Y = F + EPS(1)*F*THETA(3)*EXP(ETA_RV1) + EPS(2)*THETA(3)*EXP(ETA_RV1)'
    )
    assert (
        model.model_code.split('\n')[15]
        == '    Y = F + EPS(1)*F*EXP(ETA_RV1) + EPS(2)*EXP(ETA_RV1)'
    )


def test_set_combined_error_model_multiple_dvs(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_dvid.mod')
    model = set_combined_error_model(model, dv=1)
    rec = model.internals.control_stream.get_records('ERROR')[0]
    correct = """$ERROR
Y_1 = F + EPS(3)*F + EPS(4)
Y_2 = F + EPS(1)*F + EPS(2)

IF (DVID.EQ.1) Y = Y_1
IF (DVID.EQ.2) Y = Y_2
"""
    assert str(rec) == correct


def test_remove_error_without_f(create_model_for_test):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
CONC=A(1)/V
Y=CONC+CONC*EPS(1)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model_for_test(code)
    model = remove_error_model(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
CONC=A(1)/V
Y = CONC
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_additive_error_without_f(create_model_for_test):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
Y=CONC+CONC*EPS(1)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model_for_test(code, dataset='pheno')
    model = set_additive_error_model(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
Y = CONC + EPS(1)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA  11.2225 ; sigma
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_get_prop_init(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')

    init = _get_prop_init(model)
    assert init == 11.2225

    df = model.dataset.copy()
    df['DV'].values[:] = 0.0
    model = model.replace(dataset=df)
    init = _get_prop_init(model)
    assert init == 0.01


def test_has_additive_error_model(create_model_for_test):
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert has_additive_error_model(model)
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
CONC = THETA(1)
Y = (CONC + ETA(1)) + (CONC + ETA(1)) * ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert not has_additive_error_model(model)

    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
CONC = THETA(1)
Y = CONC + EPS(1) + CONC*EPS(2)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert not has_additive_error_model(model)

    model = create_model_for_test(code)
    assert not has_additive_error_model(model)

    code = """$PROBLEM base model
$INPUT ID DV TIME DVID
$DATA file.csv IGNORE=@

$PRED
IF (DVID.EQ.1) THEN
    Y = THETA(1) + ETA(1) + ERR(1)
ELSE
    Y = THETA(1) + THETA(1) * EPS(2)
END IF

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert has_additive_error_model(model)
    assert has_additive_error_model(model, 1)
    assert not has_additive_error_model(model, 2)


def test_has_proportional_error_model(create_model_for_test):
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert not has_proportional_error_model(model)
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
CONC = THETA(1)
Y = (CONC + ETA(1)) + (CONC + ETA(1)) * ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert has_proportional_error_model(model)

    model = create_model_for_test(code)
    assert has_proportional_error_model(model)

    code = """$PROBLEM base model
$INPUT ID DV TIME DVID
$DATA file.csv IGNORE=@

$PRED
IF (DVID.EQ.1) THEN
    Y = THETA(1) + ETA(1) + ERR(1)
ELSE
    Y = THETA(1) + THETA(1) * EPS(2)
END IF

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert not has_proportional_error_model(model)
    assert not has_proportional_error_model(model, 1)
    assert has_proportional_error_model(model, 2)


def test_has_proportional_error_model_zero_protection(testdata):
    model = read_model(testdata / 'nonmem' / 'pheno.mod')
    assert has_proportional_error_model(model)
    model = remove_error_model(model)
    model = set_proportional_error_model(model, zero_protection=True)
    assert has_proportional_error_model(model)


def test_has_combined_error_model(create_model_for_test):
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert not has_combined_error_model(model)
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
CONC = THETA(1)
Y = (CONC + ETA(1)) + (CONC + ETA(1)) * ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert not has_combined_error_model(model)

    model = create_model_for_test(code)
    assert not has_combined_error_model(model)

    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
CONC = THETA(1)
Y = (CONC + ETA(1)) + (CONC + ETA(1)) * ERR(1) + ERR(2)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$SIGMA .02
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert has_combined_error_model(model)

    code = """$PROBLEM base model
$INPUT ID DV TIME DVID
$DATA file.csv IGNORE=@

$PRED
CONC = THETA(1)
IF (DVID.EQ.1) THEN
    Y = (CONC + ETA(1)) + (CONC + ETA(1)) * ERR(1) + ERR(2)
ELSE
    Y = CONC + ERR(3)
END IF

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$SIGMA .02
$SIGMA .02
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    assert has_combined_error_model(model)
    assert has_combined_error_model(model, 1)
    assert not has_combined_error_model(model, 2)


def test_use_theta_for_error_stdev():
    code = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    model = read_model_from_string(code)
    model = use_thetas_for_error_stdev(model)
    correct = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + EPS(1)*THETA(2)

$THETA 0.1
$THETA  (0,1.0) ; SD_EPS_1
$OMEGA 0.01
$SIGMA 1 FIX
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""

    assert model.model_code == correct

    model = read_model_from_string(code)
    model = set_weighted_error_model(model)
    model = use_thetas_for_error_stdev(model)

    correct = """$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
W = THETA(2)
Y = THETA(1) + ETA(1) + EPS(1)*W

$THETA 0.1
$THETA  (0,1.0) ; SD_EPS_1
$OMEGA 0.01
$SIGMA 1 FIX
$ESTIMATION METHOD=1 INTER MAXEVALS=9990 PRINT=2 POSTHOC
"""
    assert model.model_code == correct

    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
Y=CONC+CONC*EPS(1)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = read_model_from_string(code)
    model = set_weighted_error_model(model)
    model = use_thetas_for_error_stdev(model)

    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
W = CONC*THETA(3)
Y = CONC + EPS(1)*W
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$THETA  (0,0.11506954418958998) ; SD_EPS_1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 1 FIX
$ESTIMATION METHOD=1 INTERACTION
"""

    assert model.model_code == correct


def test_set_weighted_error_model():
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
Y=CONC+CONC*EPS(1)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = read_model_from_string(code)
    model = set_weighted_error_model(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
W = CONC
Y = CONC + EPS(1)*W
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct

    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
Y=CONC+CONC*THETA(3)*EPS(1)+THETA(4)*EPS(2)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$THETA (0, 0.1) ; SD1
$THETA (0, 0.2) ; SD2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = read_model_from_string(code)
    model = set_weighted_error_model(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
W = SQRT(CONC**2*THETA(3)**2 + THETA(4)**2)
Y = CONC + EPS(1)*W
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$THETA (0, 0.1) ; SD1
$THETA (0, 0.2) ; SD2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_has_weighted_error_model():
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
Y=CONC+CONC*EPS(1)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = read_model_from_string(code)
    assert not has_weighted_error_model(model)
    model = set_weighted_error_model(model)
    assert has_weighted_error_model(model)

    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
PRED=A(1)/V
CONC=PRED
Y=CONC+CONC*THETA(3)*EPS(1)+THETA(4)*EPS(2)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$THETA (0, 0.1) ; SD1
$THETA (0, 0.2) ; SD2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = read_model_from_string(code)
    assert not has_weighted_error_model(model)
    model = set_weighted_error_model(model)
    assert has_weighted_error_model(model)


def test_set_dtbs_error_model(tmp_path, create_model_for_test):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA run1.csv IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
CONC=A(1)/V
Y=CONC+CONC*EPS(1)+EPS(2)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model_for_test(code, dataset='pheno')
    model = model.replace(name='run1')
    model = set_dtbs_error_model(model)

    with chdir(tmp_path):
        model.write_files()
        with open('run1_contr.f90') as fh:
            assert fh.readline().startswith('      subroutine contr')
        with open('run1_ccontra.f90') as fh:
            assert fh.readline().startswith('      subroutine ccontr')

    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA run1.csv IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
$ERROR
CONC=A(1)/V
W = SQRT(CONC**2*THETA(3)**2 + THETA(4)**2)
W = CONC**THETA(6)*W
IF (CONC.NE.0.AND.THETA(5).NE.0) THEN
    IPRED = (CONC**THETA(5) - 1)/THETA(5)
ELSE IF (THETA(5).EQ.0.AND.CONC.NE.0) THEN
    IPRED = LOG(CONC)
ELSE IF (CONC.EQ.0.AND.THETA(5).EQ.0) THEN
    IPRED = -1/THETA(5)
ELSE
    IPRED = -1000000000
END IF
Y = IPRED + EPS(1)*W
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$THETA  (0,0.11506954418958998) ; SD_EPS_1
$THETA  (0,0.11506954418958998) ; SD_EPS_2
$THETA  1.0 ; tbs_lambda
$THETA  0.001 ; tbs_zeta
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 1 FIX
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_set_time_varying_error_model():
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = read_model_from_string(code)
    model = set_time_varying_error_model(model, cutoff=1.0)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V
$ERROR
IF (TIME.LT.1.0) THEN
    Y = F + EPS(1)*F*THETA(3)
ELSE
    Y = F + EPS(1)*F
END IF
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$THETA  0.1 ; time_varying
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_set_time_varying_error_model_multiple_dvs(testdata, load_model_for_test):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V
$ERROR
IF (DVID.EQ.1) THEN
    Y = F + F*EPS(1)
ELSE
    Y = F + F*EPS(1) + EPS(2)
END IF
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = read_model_from_string(code)
    model = set_time_varying_error_model(model, cutoff=1.0, dv=1)

    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V
$ERROR
IF (TIME.LT.1.0) THEN
    Y_1 = F + EPS(1)*F*THETA(3)
ELSE
    Y_1 = F + EPS(1)*F
END IF
Y_2 = F + EPS(1)*F + EPS(2)
IF (DVID.EQ.1) Y = Y_1
IF (DVID.EQ.2) Y = Y_2
$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$THETA  0.1 ; time_varying
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


@pytest.mark.parametrize(
    'epsilons, same_eta, eta_names, err_ref, omega_ref',
    [
        (
            ['EPS_1'],
            False,
            None,
            'Y = F + EPS(1)*W*EXP(ETA_RV1)\n' 'IPRED=F+EPS(2)\n' 'IRES=DV-IPRED+EPS(3)\n',
            '$OMEGA  0.09 ; IIV_RUV1',
        ),
        (
            ['EPS_1', 'EPS_2'],
            False,
            None,
            'Y = F + EPS(1)*W*EXP(ETA_RV1)\n'
            'IPRED = F + EPS(2)*EXP(ETA_RV2)\n'
            'IRES=DV-IPRED+EPS(3)\n',
            '$OMEGA  0.09 ; IIV_RUV1\n' '$OMEGA  0.09 ; IIV_RUV2',
        ),
        (
            ['EPS_1', 'EPS_3'],
            False,
            None,
            'Y = F + EPS(1)*W*EXP(ETA_RV1)\n'
            'IPRED=F+EPS(2)\n'
            'IRES = DV - IPRED + EPS(3)*EXP(ETA_RV2)\n',
            '$OMEGA  0.09 ; IIV_RUV1\n' '$OMEGA  0.09 ; IIV_RUV2',
        ),
        (
            None,
            False,
            None,
            'Y = F + EPS(1)*W*EXP(ETA_RV1)\n'
            'IPRED = F + EPS(2)*EXP(ETA_RV2)\n'
            'IRES = DV - IPRED + EPS(3)*EXP(ETA_RV3)\n',
            '$OMEGA  0.09 ; IIV_RUV1\n' '$OMEGA  0.09 ; IIV_RUV2\n' '$OMEGA  0.09 ; IIV_RUV3',
        ),
        (
            None,
            True,
            None,
            'Y = F + EPS(1)*W*EXP(ETA_RV1)\n'
            'IPRED = F + EPS(2)*EXP(ETA_RV1)\n'
            'IRES = DV - IPRED + EPS(3)*EXP(ETA_RV1)\n',
            '$OMEGA  0.09 ; IIV_RUV1',
        ),
        (
            ['EPS_1'],
            False,
            ['ETA_3'],
            'Y = F + EPS(1)*W*EXP(ETA(3))\n' 'IPRED=F+EPS(2)\n' 'IRES=DV-IPRED+EPS(3)\n',
            '$OMEGA  0.09 ; IIV_RUV1',
        ),
        (
            'EPS_1',
            False,
            None,
            'Y = F + EPS(1)*W*EXP(ETA_RV1)\n' 'IPRED=F+EPS(2)\n' 'IRES=DV-IPRED+EPS(3)\n',
            '$OMEGA  0.09 ; IIV_RUV1',
        ),
    ],
)
def test_set_iiv_on_ruv(
    create_model_for_test,
    load_model_for_test,
    pheno_path,
    epsilons,
    same_eta,
    eta_names,
    err_ref,
    omega_ref,
):
    model = load_model_for_test(pheno_path)

    model_str = model.model_code
    model_more_eps = re.sub(
        'IPRED=F\nIRES=DV-IPRED', 'IPRED=F+EPS(2)\nIRES=DV-IPRED+EPS(3)', model_str
    )
    model_sigma = re.sub(
        r'\$SIGMA 0.013241', '$SIGMA 0.013241\n$SIGMA 0.1\n$SIGMA 0.1', model_more_eps
    )
    model = create_model_for_test(model_sigma)

    model = set_iiv_on_ruv(model, list_of_eps=epsilons, same_eta=same_eta, eta_names=eta_names)

    assert eta_names is None or eta_names[0] in model.random_variables.etas.names

    err_rec = model.internals.control_stream.get_records('ERROR')[0]

    assert str(err_rec) == f'$ERROR\n' f'W=F\n' f'{err_ref}' f'IWRES=IRES/W\n\n'

    omega_rec = ''.join(str(rec) for rec in model.internals.control_stream.get_records('OMEGA'))

    assert omega_rec == (
        f'$OMEGA DIAGONAL(2)\n'
        f' 0.0309626  ;       IVCL\n'
        f' 0.031128  ;        IVV\n\n'
        f'{omega_ref}\n'
    )


def test_set_iiv_on_ruv_multiple_dvs(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_dvid.mod')
    model = set_iiv_on_ruv(model, dv=1)
    rec = model.internals.control_stream.get_records('ERROR')[0]
    correct = """$ERROR
Y_1 = F + EPS(1)*F*EXP(ETA_RV1)
Y_2 = F + EPS(1)*F + EPS(2)

IF (DVID.EQ.1) Y = Y_1
IF (DVID.EQ.2) Y = Y_2
"""
    assert str(rec) == correct


def test_set_power_on_ruv_multiple_dvs(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_dvid.mod')
    updated_model = set_power_on_ruv(model, dv=1)
    rec = updated_model.internals.control_stream.get_records('ERROR')[0]
    correct = """$ERROR
Y_1 = F + EPS(1)*F**THETA(3)
Y_2 = F + EPS(1)*F + EPS(2)

IF (DVID.EQ.1) Y = Y_1
IF (DVID.EQ.2) Y = Y_2
"""
    assert str(rec) == correct

    updated_model = set_power_on_ruv(model, dv=2)
    rec = updated_model.internals.control_stream.get_records('ERROR')[0]
    correct = """$ERROR
Y_1 = F + EPS(1)*F
Y_2 = F + EPS(1)*F**THETA(3) + EPS(2)*F**THETA(4)

IF (DVID.EQ.1) Y = Y_1
IF (DVID.EQ.2) Y = Y_2
"""
    assert str(rec) == correct

    updated_model = set_power_on_ruv(model)
    rec = updated_model.internals.control_stream.get_records('ERROR')[0]
    correct = """$ERROR
Y_1 = F + EPS(1)*F**THETA(3)
Y_2 = F + EPS(1)*F**THETA(3) + EPS(2)*F**THETA(4)

IF (DVID.EQ.1) Y = Y_1
IF (DVID.EQ.2) Y = Y_2
"""
    assert str(rec) == correct

    updated_model = set_power_on_ruv(model, list_of_eps=["EPS_2"])
    rec = updated_model.internals.control_stream.get_records('ERROR')[0]
    correct = """$ERROR
Y_1 = F + EPS(1)*F
Y_2 = F + EPS(1)*F + EPS(2)*F**THETA(3)

IF (DVID.EQ.1) Y = Y_1
IF (DVID.EQ.2) Y = Y_2
"""
    assert str(rec) == correct


@pytest.mark.parametrize(
    'epsilons, err_ref, theta_ref',
    [
        (
            ['EPS_1'],
            'IF (F.EQ.0) F = 2.22500000000000E-307\n'
            'Y = F + EPS(1)*F**THETA(4)\n'
            'IPRED=F+EPS(2)\n'
            'IRES=DV-IPRED+EPS(3)',
            '$THETA  (0.01,1) ; power1',
        ),
        (
            ['EPS_1', 'EPS_2'],
            'IF (F.EQ.0) F = 2.22500000000000E-307\n'
            'Y = F + EPS(1)*F**THETA(4)\n'
            'IPRED = F + EPS(2)*F**THETA(5)\n'
            'IRES=DV-IPRED+EPS(3)',
            '$THETA  (0.01,1) ; power1\n' '$THETA  (0.01,1) ; power2',
        ),
        (
            ['EPS_1', 'EPS_3'],
            'IF (F.EQ.0) F = 2.22500000000000E-307\n'
            'Y = F + EPS(1)*F**THETA(4)\n'
            'IPRED=F+EPS(2)\n'
            'IRES = DV - IPRED + EPS(3)*F**THETA(5)',
            '$THETA  (0.01,1) ; power1\n' '$THETA  (0.01,1) ; power2',
        ),
        (
            None,
            'IF (F.EQ.0) F = 2.22500000000000E-307\n'
            'Y = F + EPS(1)*F**THETA(4)\n'
            'IPRED = F + EPS(2)*F**THETA(5)\n'
            'IRES = DV - IPRED + EPS(3)*F**THETA(6)',
            '$THETA  (0.01,1) ; power1\n' '$THETA  (0.01,1) ; power2\n' '$THETA  (0.01,1) ; power3',
        ),
    ],
)
def test_set_power_on_ruv(
    load_model_for_test, create_model_for_test, testdata, epsilons, err_ref, theta_ref, tmp_path
):
    shutil.copy(testdata / 'nonmem/pheno_real.mod', tmp_path / 'run1.mod')
    shutil.copy(testdata / 'nonmem/pheno_real.phi', tmp_path / 'run1.phi')
    shutil.copy(testdata / 'nonmem/pheno_real.ext', tmp_path / 'run1.ext')
    shutil.copy(testdata / 'nonmem/pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        model_pheno = load_model_for_test('run1.mod')
        model_more_eps = re.sub(
            r'V=TVV\*EXP\(ETA\(2\)\)',
            'V=TVV',
            model_pheno.model_code,
        )
        model_more_eps = re.sub(
            r'( 0.031128  ;        IVV\n)',
            '$SIGMA 0.1\n$SIGMA 0.1',
            model_more_eps,
        )
        model_more_eps = re.sub(
            r'IPRED=F\nIRES=DV-IPRED',
            r'IPRED=F+EPS(2)\nIRES=DV-IPRED+EPS(3)',
            model_more_eps,
        )
        model = create_model_for_test(model_more_eps)
        model = model.replace(dataset=model_pheno.dataset)

        model = set_power_on_ruv(model, epsilons, zero_protection=True)

        rec_err = str(model.internals.control_stream.get_records('ERROR')[0])
        correct = f'$ERROR\n' f'W=F\n' f'{err_ref}\n' f'IWRES=IRES/W\n\n'
        assert rec_err == correct

        rec_theta = ''.join(str(rec) for rec in model.internals.control_stream.get_records('THETA'))

        assert (
            rec_theta == f'$THETA (0,0.00469307) ; PTVCL\n'
            f'$THETA (0,1.00916) ; PTVV\n'
            f'$THETA (-.99,.1)\n'
            f'{theta_ref}\n'
        )


def test_set_power_on_ruv_with_zero_protect(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = remove_error_model(model)
    model = set_proportional_error_model(model)
    model = set_power_on_ruv(model)

    assert 'F + EPS(1)*IPREDADJ**THETA(3)' in model.model_code



from pharmpy.deps import sympy
from pharmpy.model import Model, Parameters, RandomVariables


def replace_non_random_rvs(model: Model):
    """Replace all random variables that are not actually random

    Some random variables are constant. For example a normal
    distribution with the variance parameter fixed to 0 will always
    yield a single value when sampled. This function will find all such
    random variables and replace them with their constant value in the model.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    Model
        A new model
    """

    dists = model.random_variables
    keep = []  # Dists
    remove = []  # Pop-params
    d = {}

    for dist in dists:
        for parname in dist.parameter_names:
            param = model.parameters[parname]
            if not (param.init == 0.0 and param.fix):
                keep.append(dist)
                break
        else:
            for parname in dist.parameter_names:
                remove.append(parname)
                d[sympy.Symbol(parname)] = sympy.Integer(0)
            for name in dist.names:
                d[sympy.Symbol(name)] = sympy.Integer(0)

    new_parameters = Parameters(tuple(p for p in model.parameters if p.name not in remove))
    new_statements = model.statements.subs(d)
    new_rvs = RandomVariables(
        keep, eta_levels=dists.eta_levels, epsilon_levels=dists.epsilon_levels
    )

    model = model.replace(
        statements=new_statements, parameters=new_parameters, random_variables=new_rvs
    )
    model = model.update_source()
    return model



import math
import re
from typing import Union, cast

from pharmpy.internals.parse import AttrToken, AttrTree
from pharmpy.internals.parse.generic import eval_token, remove_token_and_space
from pharmpy.model import ModelSyntaxError

from .record import Record

MAX_UPPER_BOUND = 1000000
MIN_LOWER_BOUND = -1000000
INF = float("inf")


def lower_token(theta):
    # Get raw lower bound for a theta subtree
    # Either a float, None for non-existing or 'neginf' for -INF
    if theta.find('low'):
        low = theta.subtree('low')
        if low.find('NEG_INF'):
            lower = 'neginf'
        else:
            lower = eval_token(next(iter(low.tokens)))
    else:
        lower = None
    return lower


def upper_token(theta):
    # Get raw upper bound for a theta subtree
    # Either a float, None for non-existing or 'inf' for INF
    if theta.find('up'):
        up = theta.subtree('up')
        if up.find('POS_INF'):
            upper = 'inf'
        else:
            upper = eval_token(next(iter(up.tokens)))
    else:
        upper = None
    return upper


class ThetaRecord(Record):
    def _multiple(self, theta: AttrTree) -> int:
        """Return the multiple (xn) of a theta or 1 if no multiple"""
        if theta.find('n'):
            return cast(int, eval_token(theta.subtree('n').leaf('INT')))
        else:
            return 1

    @property
    def inits(self):
        # List of initial estimates for all thetas
        inits = []
        for theta in self.root.subtrees('theta'):
            init = eval_token(theta.subtree('init').leaf('NUMERIC'))
            if init == MAX_UPPER_BOUND or init == MIN_LOWER_BOUND:
                raise ModelSyntaxError(
                    f"Initial estimate of THETA cannot be {MAX_UPPER_BOUND} or {MIN_LOWER_BOUND}"
                )
            n = self._multiple(theta)
            inits.extend([init] * n)
        return inits

    @property
    def fixs(self):
        # List of fixedness for all thetas
        fixs = []
        for theta in self.root.subtrees('theta'):
            init = eval_token(theta.subtree('init').leaf('NUMERIC'))
            lowtok = lower_token(theta)
            uptok = upper_token(theta)
            # Raise if FIX is within the parentheses and explicit bounds
            # are used that are not the same as the init
            inparens = False
            fix = False
            for child in theta.children:
                if child.rule == 'LPAR':
                    inparens = True
                elif child.rule == 'RPAR':
                    inparens = False
                elif child.rule == 'FIX':
                    if inparens:
                        if lowtok is not None:
                            message = (
                                "FIX inside parentheses of $THETA requires all bounds to be"
                                " the same as the initial value. "
                            )
                            if uptok is not None:
                                if not (lowtok == uptok == init):
                                    raise ModelSyntaxError(
                                        f"{message}init={init}, lower={lowtok}, upper={uptok}"
                                    )
                            else:
                                if not (lowtok == init):
                                    raise ModelSyntaxError(f"{message}init={init}, lower={lowtok}")
                    fix = True
                    break

            if not fix and uptok is None and lowtok == init:
                raise ModelSyntaxError(
                    "Lower bound cannot be equal to initial estimate of THETA unless FIX"
                )
            if not fix and init == 0:
                raise ModelSyntaxError("Initial estimate of THETA cannot be 0 unless fixed")
            n = self._multiple(theta)
            fixs.extend([fix] * n)
        return fixs

    @property
    def bounds(self):
        # List of tuples of lower, upper bounds for all thetas
        bounds = []
        for theta in self.root.subtrees('theta'):
            lowtok = lower_token(theta)
            if isinstance(lowtok, float):
                if lowtok == MIN_LOWER_BOUND:
                    lower = -INF
                elif lowtok < MIN_LOWER_BOUND:
                    raise ModelSyntaxError(f"Too low lower bound: {lowtok} < {MIN_LOWER_BOUND}")
                else:
                    lower = lowtok
            else:
                lower = -INF

            uptok = upper_token(theta)
            if isinstance(uptok, float):
                if uptok == MAX_UPPER_BOUND:
                    upper = INF
                elif uptok > MAX_UPPER_BOUND:
                    raise ModelSyntaxError(f"Too high upper bound: {uptok} > {MAX_UPPER_BOUND}")
                else:
                    upper = uptok
            else:
                upper = INF

            n = self._multiple(theta)
            bounds.extend([(lower, upper)] * n)
        return bounds

    @property
    def comment_names(self):
        # A list of all comment names in the record
        # None if missing
        names = []
        intheta = False
        n = 0
        for node in self.root.tree_walk():
            if node.rule == 'theta':
                intheta = True
                if n != 0:
                    names.extend([None] * n)
                n = self._multiple(node)
            if intheta and node.rule == 'COMMENT':
                m = re.search(r';\s*([a-zA-Z_]\w*)', str(node))
                if m:
                    names.append(m.group(1))
                else:
                    names.append(None)
                n -= 1
                if n == 0:
                    intheta = False

        if n != 0:
            names.extend([None] * n)

        return names

    def update(self, parameters):
        """From a Parameters update the THETAs in this record

        parameters - Only parameters defined by this $THETA record
        """
        i = 0

        def _update_theta(child: Union[AttrTree, AttrToken]):
            nonlocal i

            if not isinstance(child, AttrTree) or not child.rule == 'theta':
                return child

            theta = child

            param = parameters[i]
            new_init = param.init
            init = theta.subtree('init')
            if eval_token(init.leaf('NUMERIC')) != new_init:
                init = init.replace_first(AttrToken('NUMERIC', str(new_init)))
            theta = theta.replace_first(init)
            fix = bool(theta.find('FIX'))
            if fix != param.fix:
                if param.fix:
                    space = AttrToken('WS', ' ')
                    fix_token = AttrToken('FIX', 'FIX')
                    theta = AttrTree(theta.rule, theta.children + (space, fix_token))
                else:
                    theta = remove_token_and_space(theta, 'FIX')

            up = theta.find('up')
            if up != param.upper:
                if up is None and param.upper < 1000000:
                    comma = AttrToken('COMMA', ',')
                    upper = AttrToken('NUMERIC', param.upper)
                    up_node = AttrTree('up', (upper,))
                    keep = []
                    for node in theta.children:
                        keep.append(node)
                        if node.rule == 'init':
                            keep.extend((comma, up_node))
                    theta = AttrTree(theta.rule, tuple(keep))

            low = theta.find('low')
            if low != param.lower:
                comma = AttrToken('COMMA', ',')
                if not math.isinf(param.lower) and int(param.lower) == param.lower:
                    lower = AttrToken('NUMERIC', int(param.lower))
                else:
                    lower = AttrToken('NUMERIC', param.lower)
                low_node = AttrTree('low', (lower,))
                keep = []
                for node in theta.children:
                    keep.append(node)
                    if node.rule == 'low':
                        keep.remove(node)
                        keep.extend((low_node,))
                theta = AttrTree(theta.rule, tuple(keep))

            i += self._multiple(theta)

            return theta

        new_tree = self.root.map(_update_theta)
        return ThetaRecord(self.name, self.raw_name, new_tree)

    def remove(self, inds):
        if not inds:
            return self
        keep = []
        i = 0
        for node in self.root.children:
            if node.rule == 'theta':
                if i not in inds:
                    keep.append(node)
                i += 1
            else:
                keep.append(node)

        new_tree = AttrTree(self.root.rule, tuple(keep))
        return ThetaRecord(self.name, self.raw_name, new_tree)

    def __len__(self):
        """Number of thetas in this record"""
        tot = 0
        for theta in self.root.subtrees('theta'):
            tot += self._multiple(theta)
        return tot



from typing import Union

from pharmpy.deps import numpy as np
from pharmpy.modeling import calculate_aic, calculate_bic

# All functions used for comparing a set of candidate models and
# ranking them take the same base arguments
# base - the base model
# candidates - a list of candidate models
# returns a list which is a subset of the candidates with the best ranked at the beginning
# models can be removed from candidates if they don't pass some criteria


def ofv(base, candidates, cutoff=3.84):
    return rank_models('ofv', base, candidates, cutoff)


def aic(base, candidates, cutoff=None):
    return rank_models('aic', base, candidates, cutoff)


def bic(base, candidates, cutoff=None, bic_type: Union[None, str] = 'mixed'):
    return rank_models('bic', base, candidates, cutoff, bic_type=bic_type)


def rank_models(rank_type, base, candidates, cutoff=None, bic_type: Union[None, str] = 'mixed'):
    diff_dict = _create_diff_dict(rank_type, base, candidates, bic_type)
    if cutoff is None:
        filtered = [model for model in candidates if model.modelfit_results is not None]
    else:
        filtered = [model for model in candidates if diff_dict[model.name] >= cutoff]

    def fn(model):
        if rank_type == 'aic':
            return calculate_aic(model, model.modelfit_results.ofv)
        elif rank_type == 'bic':
            return calculate_bic(model, model.modelfit_results.ofv, bic_type)
        else:
            return model.modelfit_results.ofv

    srtd = sorted(filtered, key=fn)  # FIXME: if same rankfunc value?
    return srtd, diff_dict


def _create_diff_dict(rank_type, base, candidates, bic_type):
    diff_dict = {}
    for model in candidates:
        # FIXME: Way to handle if start model fails
        if base.modelfit_results is None or model.modelfit_results is None:
            diff = np.nan
        elif rank_type == 'aic':
            diff = calculate_aic(base, base.modelfit_results.ofv) - calculate_aic(
                model, model.modelfit_results.ofv
            )
        elif rank_type == 'bic':
            diff = calculate_bic(base, base.modelfit_results.ofv, bic_type) - calculate_bic(
                model, model.modelfit_results.ofv, bic_type
            )
        else:
            diff = base.modelfit_results.ofv - model.modelfit_results.ofv
        diff_dict[model.name] = diff
    return diff_dict



import os
import sys

sys.path.append(os.path.abspath('./_ext'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.githubpages',
    'sphinx.ext.graphviz',
    'sphinx.ext.ifconfig',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_automodapi.automodapi',
    'jupyter_sphinx',
    'sphinx_tabs.tabs',
    'sphinxcontrib.autoprogram',
    'pharmpy_snippet',
]
if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_US'

source_suffix = '.rst'
master_doc = 'index'
project = 'Pharmpy'
year = '2018-2024'
authors = ['the Pharmpy development team']
copyright = '{0}; {1}'.format(year, ', '.join(authors))
version = release = '0.106.0'
html_show_sourcelink = False

pygments_style = 'trac'
templates_path = ['.']

html_static_path = ['.']
html_theme = "pydata_sphinx_theme"
html_css_files = [
    'custom.css',
]
html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = False
html_theme_options = {
    'globaltoc_maxdepth': 2,
    "logo": {
        "image_light": "Pharmpy_logo.svg",
        "image_dark": "Pharmpy_logo_dark.svg",
    },
    'github_url': 'https://github.com/pharmpy/pharmpy',
    "icon_links": [
        {
            "name": "Report issues",
            "url": "https://github.com/pharmpy/pharmpy/issues",
            "icon": "fa fa-comments",
        }
    ]
}
html_favicon = 'images/Pharmpy_symbol.svg'
html_context = {
   "default_mode": "light"
}
html_short_title = '%s-%s' % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False

automodapi_inheritance_diagram = False
autodoc_typehints = 'none'
graphviz_output_format = 'svg'

import doctest
doctest_default_flags = doctest.NORMALIZE_WHITESPACE
doctest_global_setup = '''
import pandas as pd
pd.set_option('display.width', 1000)
pd.set_option('display.max_columns', 1000)
'''

# Page is reachable from browser but gives 403 with GET
linkcheck_ignore = [r'https://doi.org/10.1002/psp4.12741']



from typing import TYPE_CHECKING, Union

from ..unicode import int_to_superscript
from .parse import parse as parse_expr
from .subs import subs

if TYPE_CHECKING:
    import sympy
    import sympy.printing as sympy_printing
else:
    from pharmpy.deps import sympy, sympy_printing


def parse(s: Union[str, int, float, sympy.Expr]) -> sympy.Expr:
    if not isinstance(s, sympy.Expr):
        return subs(parse_expr(s), _unit_subs(), simultaneous=True)
    else:
        return s
    # return (
    #    subs(parse_expr(s), _unit_subs(), simultaneous=True) if not isinstance(s, sympy.Expr) else s
    # )


_unit_subs_cache = None


def _unit_subs():
    global _unit_subs_cache
    if _unit_subs_cache is None:
        subs = {}
        import sympy.physics.units as units

        for k, v in units.__dict__.items():
            if isinstance(v, sympy.Expr) and v.has(units.Unit):
                subs[sympy.Symbol(k)] = v

        _unit_subs_cache = subs

    return _unit_subs_cache


# sympy_printing.str.StrPrinter._default_settings['abbrev'] = True


class UnitPrinter(sympy_printing.str.StrPrinter):
    """Print physical unit as unicode"""

    def _print_Mul(self, expr):
        pow_strings = [self._print(e) for e in expr.args if e.is_Pow]
        plain_strings = [self._print(e) for e in expr.args if not e.is_Pow]
        all_strings = sorted(plain_strings) + sorted(pow_strings)
        return '⋅'.join(all_strings)

    def _print_Pow(self, expr, rational=False):
        base = expr.args[0]
        exp = expr.args[1]
        if exp.is_Integer:
            exp_str = int_to_superscript(int(exp))
        else:
            exp_str = "^" + self._print(exp)
        return self._print(base) + exp_str

    def _print_Quantity(self, expr):
        # FIXME: sympy cannot handle the abbreviation of ml
        if str(expr) == "milliliter":
            return "ml"
        else:
            return str(expr.args[1])


def unit_string(expr: sympy.Expr) -> str:
    printer = UnitPrinter()
    return printer._print(expr)



from __future__ import annotations

import re
from io import StringIO
from pathlib import Path
from typing import List, Optional, Union

from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.internals.math import flattened_to_symmetric


class NONMEMTableFile:
    """A NONMEM table file that can contain multiple tables"""

    def __init__(
        self,
        path: Optional[Union[str, Path]] = None,
        tables: Optional[List[NONMEMTable]] = None,
        notitle: bool = False,
        nolabel: bool = False,
    ):
        if path is not None:
            path = Path(path)
            suffix = path.suffix
            tables = []
            if path.stat().st_size == 0:
                raise OSError("Empty table file")
            with open(str(path), 'r') as tablefile:
                if notitle:
                    table = self._parse_table(
                        tablefile.read().splitlines(keepends=True), notitle=notitle, nolabel=nolabel
                    )
                    tables.append(table)
                else:
                    current = []
                    for line in tablefile:
                        if line.startswith("TABLE NO."):
                            if current:
                                table = self._parse_table(current, suffix)
                                tables.append(table)
                            current = [line]
                        else:
                            current.append(line)
                    table = self._parse_table(current, suffix)
                    tables.append(table)
            self.tables = tables
        elif tables is not None:
            self.tables = tables
        else:
            raise ValueError('NONMEMTableFile: path and tables cannot be both None')

    def _parse_table(
        self,
        content: List[str],
        suffix: Optional[str] = None,
        notitle: bool = False,
        nolabel: bool = False,
    ) -> NONMEMTable:
        # NOTE: Content lines must contain endlines!

        table_line = None if notitle else content.pop(0)

        if suffix == '.ext':
            table = ExtTable(''.join(content))
        elif suffix == '.phi':
            table = PhiTable(''.join(content))
        elif suffix == '.cov' or suffix == '.cor' or suffix == '.coi':
            table = CovTable(''.join(content))
        else:
            # Remove repeated header lines, but not the first
            content[1:] = [line for line in content[1:] if not re.match(r'\s[A-Za-z_]', line)]
            table = NONMEMTable(''.join(content))  # Fallback to non-specific table type

        if table_line is not None:
            m = re.match(r'TABLE NO.\s+(\d+)', table_line)
            if not m:
                raise ValueError(f"Illegal {suffix}-file: missing TABLE NO.")
            table.number = int(m.group(1))
            table.is_evaluation = False
            if re.search(r'(Evaluation)', table_line):
                table.is_evaluation = True  # No estimation step was run

            m = re.match(
                r'TABLE NO.\s+\d+: (.*?)(?:: ([\w-]+))?: (?:Goal Function=(.*): )?Problem=(\d+) '
                r'Subproblem=(\d+) Superproblem1=(\d+) Iteration1=(\d+) Superproblem2=(\d+) '
                r'Iteration2=(\d+)',
                table_line,
            )

            if m:
                table.method = m.group(1)
                table.design_optimality = m.group(2)
                table.goal_function = m.group(3)
                table.problem = int(m.group(4))
                table.subproblem = int(m.group(5))
                table.superproblem1 = int(m.group(6))
                table.iteration1 = int(m.group(7))
                table.superproblem2 = int(m.group(8))
                table.iteration2 = int(m.group(9))

        return table

    def __iter__(self):
        return iter(self.tables)

    def __getitem__(self, i):
        return self.tables[i]

    def __len__(self):
        return len(self.tables)

    @property
    def table(
        self, problem=1, subproblem=0, superproblem1=0, iteration1=0, superproblem2=0, iteration2=0
    ):
        for t in self.tables:
            if (
                t.problem == problem
                and t.subproblem == subproblem
                and t.superproblem1 == superproblem1
                and t.iteration1 == iteration1
                and t.superproblem2 == superproblem2
                and t.iteration2 == iteration2
            ):
                return t

    def table_no(self, table_number=1):
        for table in self.tables:
            if table.number == table_number:
                return table

    def write(self, path):
        with open(path, 'w') as df:
            for table in self.tables:
                print('TABLE NO.     1', file=df)
                print(table.content, file=df, end='')


class NONMEMTable:
    """A NONMEM output table."""

    number: Optional[int] = None
    is_evaluation: Optional[bool] = None
    method: Optional[str] = None
    design_optimality: Optional[str] = None
    goal_function: Optional[str] = None
    problem: Optional[int] = None
    subproblem: Optional[int] = None
    superproblem1: Optional[int] = None
    iteration1: Optional[int] = None
    superproblem2: Optional[int] = None
    iteration2: Optional[int] = None

    def __init__(self, content=None, df=None):
        if content is not None:
            self._df = pd.read_table(StringIO(content), sep=r'\s+', engine='c')
        elif df is not None:
            self._df = df
        else:
            raise ValueError('NONMEMTable: content and df cannot be both None')

    @property
    def data_frame(self):
        return self._df

    @property
    def content(self):
        df = self._df.copy(deep=True)
        df.reset_index(inplace=True)
        df.insert(loc=0, column='SUBJECT_NO', value=np.arange(len(df)) + 1)
        fmt = '%13d%13d' + '%13.5E' * (len(df.columns) - 2)

        header_fmt = ' %-12s' * len(df.columns) + '\n'
        header = header_fmt % tuple(df.columns)

        with StringIO() as s:
            np.savetxt(s, df.values, fmt=fmt)
            body = s.getvalue()

        return header + body

    @staticmethod
    def rename_index(df, ext=True):
        """If columns rename also the row index"""
        theta_labels = [x for x in df.columns if x.startswith('THETA')]
        omega_labels = [x for x in df.columns if x.startswith('OMEGA')]
        sigma_labels = [x for x in df.columns if x.startswith('SIGMA')]
        labels = theta_labels + omega_labels + sigma_labels
        if ext:
            labels = ['ITERATION'] + labels + ['OBJ']
        else:
            df = df.reindex(labels)
        df = df.reindex(labels, axis=1)
        df.columns = df.columns.str.replace(r'THETA(\d+)', r'THETA(\1)', regex=True)
        if not ext:
            df.index = df.index.str.replace(r'THETA(\d+)', r'THETA(\1)', regex=True)
        return df


class CovTable(NONMEMTable):
    @property
    def data_frame(self):
        df = self._df.copy(deep=True)
        df.set_index('NAME', inplace=True)
        df.index.name = None
        df = NONMEMTable.rename_index(df, ext=False)
        df = df.loc[(df != 0).any(axis=1), (df != 0).any(axis=0)]  # Remove FIX
        return df


class PhiTable(NONMEMTable):
    @property
    def iofv(self):
        df = self._df.copy(deep=True)
        # Remove all zero lines. These are from individuals with no observations
        df = df.loc[df.iloc[:, 2:].any(axis=1)]
        iofv = df[['ID', 'OBJ']]
        iofv.set_index('ID', inplace=True)
        iofv.columns = ['iOFV']
        return iofv['iOFV']

    @property
    def etas(self):
        df = self._df
        df = df.loc[df.iloc[:, 2:].any(axis=1)]
        eta_col_names = [col for col in df if col.startswith('ETA') or col.startswith('PHI')]
        etas = df[eta_col_names]
        etas.index = df['ID']
        return etas

    @property
    def etcs(self):
        ids, eta_col_names, matrix_array = self.etc_data()
        etc_frames = [
            pd.DataFrame(matrix, columns=eta_col_names, index=eta_col_names)
            for matrix in matrix_array
        ]
        etcs = pd.Series(etc_frames, index=ids, dtype='object')
        return etcs

    def etc_data(self):
        # Note that PHC == ETC
        df = self._df
        df = df.loc[df.iloc[:, 2:].any(axis=1)]
        eta_col_names = [col for col in df if col.startswith('ETA') or col.startswith('PHI')]
        etc_col_names = [col for col in df if col.startswith('ETC') or col.startswith('PHC')]
        vals = df[etc_col_names].values
        matrix_array = [flattened_to_symmetric(x) for x in vals]
        colnames = [f'ETA{name[3:]}' for name in eta_col_names]
        return df['ID'], colnames, matrix_array


class ExtTable(NONMEMTable):
    @property
    def data_frame(self):
        df = self._df.copy(deep=True)
        df = NONMEMTable.rename_index(df)
        if df['ITERATION'].isnull().values.all():
            raise ValueError('Broken table in ext-file')
        return df

    def _get_parameters(self, iteration, include_thetas=True):
        df = self.data_frame
        row = df.loc[df['ITERATION'] == iteration]
        if df['ITERATION'].isnull().values.all():
            return np.nan
        elif row.empty:
            raise KeyError(f'Row {iteration} not available in ext-file')
        del row['ITERATION']
        del row['OBJ']
        if not include_thetas:
            row = row[row.columns.drop(list(row.filter(regex='THETA')))]
        return row.squeeze()

    def _get_ofv(self, iteration):
        df = self.data_frame
        row = df.loc[df['ITERATION'] == iteration]
        if df['ITERATION'].isnull().values.all():
            return np.nan
        elif row.empty:
            raise KeyError(f'Row {iteration} not available in ext-file')
        row.squeeze()
        return row['OBJ'].squeeze()

    @property
    def iterations(self):
        """List of all iteration numbers (except the special negative)"""
        it = self.data_frame['ITERATION']
        return list(it[it >= 0])

    @property
    def final_parameter_estimates(self):
        """Get the final parameter estimates as a Series
        A specific parameter estimate can be retreived as
        final_parameter_estimates['THETA1']
        Fallback to estimates from last iteration in case
        of an aborted or failed run.
        """
        try:
            ser = self._get_parameters(-1000000000)
        except KeyError:
            final_iter = max(self.iterations)
            ser = self._get_parameters(final_iter)
        ser.name = 'estimates'
        return ser

    @property
    def standard_errors(self):
        """Get the standard errors of the parameter estimates as a Series"""
        ser = self._get_parameters(-1000000001)
        ser.name = 'SE'
        return ser

    @property
    def condition_number(self):
        """Get the condition number of the correlation matrix"""
        ser = self._get_parameters(-1000000003)
        return ser.values[0]

    @property
    def omega_sigma_stdcorr(self):
        """Get the omegas and sigmas in standard deviation/correlation form"""
        return self._get_parameters(-1000000004, include_thetas=False)

    @property
    def omega_sigma_se_stdcorr(self):
        """The standard error of the omegas and sigmas in stdcorr form"""
        return self._get_parameters(-1000000005, include_thetas=False)

    @property
    def fixed(self):
        """What parameters were fixed?"""
        fix = self._get_parameters(-1000000006)
        return fix.apply(bool)

    @property
    def final_ofv(self):
        try:
            ser = self._get_ofv(-1000000000)
        except KeyError:
            final_iter = max(self.iterations)
            ser = self._get_ofv(final_iter)
        return ser

    @property
    def initial_ofv(self):
        try:
            ser = self._get_ofv(0)
        except KeyError:
            ser = self._get_ofv(-1000000000)
        return ser



from typing import Dict, Set, TypeVar

T = TypeVar('T')
U = TypeVar('U')


def inverse(g: Dict[T, Set[U]]) -> Dict[U, Set[T]]:
    h = {}

    for left, deps in g.items():
        for right in deps:
            if right in h:
                h[right].add(left)
            else:
                h[right] = {left}

    return h



import pytest

from pharmpy.modeling import (
    add_iiv,
    add_individual_parameter,
    add_peripheral_compartment,
    add_pk_iiv,
    create_joint_distribution,
    fix_parameters,
)
from pharmpy.tools import read_modelfit_results
from pharmpy.tools.iivsearch.algorithms import (
    _create_param_dict,
    _is_rv_block_structure,
    _rv_block_structures,
    brute_force_block_structure,
    brute_force_no_of_etas,
    create_eta_blocks,
)
from pharmpy.tools.iivsearch.tool import create_workflow, validate_input
from pharmpy.workflows import Workflow


@pytest.mark.parametrize(
    'list_of_parameters, expected_values',
    [([], 4), (['IVCL'], 1), (["IVCL", "IVV"], 0)],
)
def test_brute_force_block_structure_ignore_fixed_params(
    load_model_for_test, testdata, list_of_parameters, expected_values
):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    model = add_individual_parameter(model, 'PD1')
    model = add_iiv(model, 'PD1', 'exp')
    model = fix_parameters(model, list_of_parameters)
    wf = brute_force_block_structure(model)
    fit_tasks = [task.name for task in wf.tasks if task.name.startswith('run')]
    assert len(fit_tasks) == expected_values


@pytest.mark.parametrize(
    'list_of_parameters, expected_values',
    [([], 3), (['CL'], 1), (["CL", "V"], 0)],
)
def test_brute_force_no_of_etas_keep(
    load_model_for_test, testdata, list_of_parameters, expected_values
):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    wf = brute_force_no_of_etas(model, keep=list_of_parameters)
    fit_tasks = [task.name for task in wf.tasks if task.name.startswith('run')]
    assert len(fit_tasks) == expected_values


def test_brute_force_no_of_etas_fixed(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    model = fix_parameters(model, 'IVCL')
    wf = brute_force_no_of_etas(model)
    fit_tasks = [task.name for task in wf.tasks if task.name.startswith('run')]
    assert len(fit_tasks) == 1


@pytest.mark.parametrize(
    'list_of_parameters, no_of_models',
    [([], 7), (['QP1'], 15)],
)
def test_brute_force_no_of_etas(load_model_for_test, testdata, list_of_parameters, no_of_models):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'mox2.mod')
    model = add_peripheral_compartment(model)
    model = add_iiv(model, list_of_parameters, 'add')
    wf = brute_force_no_of_etas(model)
    fit_tasks = [task.name for task in wf.tasks if task.name.startswith('run')]

    assert len(fit_tasks) == no_of_models


@pytest.mark.parametrize(
    'list_of_parameters, block_structure, no_of_models',
    [([], [], 4), (['QP1'], [], 14), ([], ['ETA_1', 'ETA_2'], 4)],
)
def test_brute_force_block_structure(
    load_model_for_test, testdata, list_of_parameters, block_structure, no_of_models
):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'mox2.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'models' / 'mox2.mod')
    model = add_peripheral_compartment(model)
    model = add_iiv(model, list_of_parameters, 'add')
    if block_structure:
        model = create_joint_distribution(
            model, block_structure, individual_estimates=res.individual_estimates
        )

    wf = brute_force_block_structure(model)
    fit_tasks = [task.name for task in wf.tasks if task.name.startswith('run')]

    assert len(fit_tasks) == no_of_models


def test_rv_block_structures_4_etas(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    model = add_iiv(model, ['TAD', 'S1'], 'exp')

    block_structures = list(_rv_block_structures(model.random_variables.iiv.names))

    assert len(block_structures) == 15

    block_structures_integer_partitions = [
        tuple(map(len, block_structure)) for block_structure in block_structures
    ]
    assert block_structures_integer_partitions.count((4,)) == 1
    assert block_structures_integer_partitions.count((1, 3)) == 4
    assert block_structures_integer_partitions.count((2, 2)) == 3
    assert block_structures_integer_partitions.count((1, 1, 2)) == 6
    assert block_structures_integer_partitions.count((1, 1, 1, 1)) == 1


def test_rv_block_structures_5_etas(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    model = add_iiv(model, ['TVCL', 'TAD', 'S1'], 'exp')

    block_structures = list(_rv_block_structures(model.random_variables.iiv.names))
    assert len(block_structures) == 52

    block_structures_integer_partitions = [
        tuple(map(len, block_structure)) for block_structure in block_structures
    ]
    assert block_structures_integer_partitions.count((5,)) == 1
    assert block_structures_integer_partitions.count((1, 4)) == 5
    assert block_structures_integer_partitions.count((2, 3)) == 10
    assert block_structures_integer_partitions.count((1, 1, 3)) == 10
    assert block_structures_integer_partitions.count((1, 2, 2)) == 15
    assert block_structures_integer_partitions.count((1, 1, 1, 2)) == 10
    assert block_structures_integer_partitions.count((1, 1, 1, 1, 1)) == 1


def test_is_rv_block_structure(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    res = read_modelfit_results(pheno_path)
    model = add_iiv(model, ['TAD', 'S1'], 'exp')

    etas_block_structure = (('ETA_1', 'ETA_2'), ('ETA_TAD',), ('ETA_S1',))
    model = create_joint_distribution(
        model,
        list(etas_block_structure[0]),
        individual_estimates=res.individual_estimates,
    )
    etas = model.random_variables.iiv
    assert _is_rv_block_structure(etas, etas_block_structure, [])

    etas_block_structure = (('ETA_1',), ('ETA_2',), ('ETA_TAD', 'ETA_S1'))
    assert not _is_rv_block_structure(etas, etas_block_structure, [])

    etas_block_structure = (('ETA_1',), ('ETA_2', 'ETA_TAD'), ('ETA_S1',))
    assert not _is_rv_block_structure(etas, etas_block_structure, [])

    model = create_joint_distribution(model, individual_estimates=res.individual_estimates)
    etas_block_structure = (('ETA_1', 'ETA_2', 'ETA_TAD', 'ETA_S1'),)
    etas = model.random_variables.iiv
    assert _is_rv_block_structure(etas, etas_block_structure, [])


def test_create_joint_dist(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'mox2.mod')
    res = read_modelfit_results(testdata / 'nonmem' / 'models' / 'mox2.mod')
    # FIXME!!
    model = model.replace(modelfit_results=res)

    model = add_peripheral_compartment(model)
    model = add_pk_iiv(model)
    etas_block_structure = (('ETA_1', 'ETA_2'), ('ETA_QP1',), ('ETA_VP1',))
    model = create_eta_blocks(etas_block_structure, model)
    assert len(model.random_variables.iiv) == 4

    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'mox2.mod')
    model = model.replace(modelfit_results=res)
    model = add_peripheral_compartment(model)
    model = add_pk_iiv(model)
    model = create_joint_distribution(
        model,
        ['ETA_1', 'ETA_2'],
        individual_estimates=res.individual_estimates,
    )
    etas_block_structure = (('ETA_1',), ('ETA_2',), ('ETA_3', 'ETA_VP1', 'ETA_QP1'))
    model = create_eta_blocks(etas_block_structure, model)
    assert len(model.random_variables.iiv) == 3


def test_get_param_names(create_model_for_test, load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'mox2.mod')

    param_dict = _create_param_dict(model, model.random_variables.iiv)
    param_dict_ref = {'ETA_1': 'CL', 'ETA_2': 'VC', 'ETA_3': 'MAT'}

    assert param_dict == param_dict_ref

    model_code = model.model_code.replace(
        'CL = THETA(1) * EXP(ETA(1))', 'ETA_1 = ETA(1)\nCL = THETA(1) * EXP(ETA_1)'
    )
    model = create_model_for_test(model_code)

    param_dict = _create_param_dict(model, model.random_variables.iiv)

    assert param_dict == param_dict_ref


def test_create_workflow():
    assert isinstance(create_workflow('brute_force'), Workflow)


def test_create_workflow_with_model(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    assert isinstance(create_workflow('brute_force', model=model), Workflow)


def test_validate_input():
    validate_input('brute_force')


def test_validate_input_with_model(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    validate_input('brute_force', model=model)


@pytest.mark.parametrize(
    ('model_path', 'arguments', 'exception', 'match'),
    [
        (None, dict(algorithm=1), ValueError, 'Invalid `algorithm`'),
        (None, dict(algorithm='brute_force_no_of_eta'), ValueError, 'Invalid `algorithm`'),
        (None, dict(rank_type=1), ValueError, 'Invalid `rank_type`'),
        (None, dict(rank_type='bi'), ValueError, 'Invalid `rank_type`'),
        (None, dict(iiv_strategy=['no_add']), ValueError, 'Invalid `iiv_strategy`'),
        (None, dict(iiv_strategy='diagonal'), ValueError, 'Invalid `iiv_strategy`'),
        (None, dict(cutoff='1'), TypeError, 'Invalid `cutoff`'),
        (
            None,
            dict(model=1),
            TypeError,
            'Invalid `model`',
        ),
    ],
)
def test_validate_input_raises(
    load_model_for_test,
    testdata,
    model_path,
    arguments,
    exception,
    match,
):
    model = load_model_for_test(testdata.joinpath(*model_path)) if model_path else None

    harmless_arguments = dict(
        algorithm='brute_force',
    )

    kwargs = {**harmless_arguments, 'model': model, **arguments}

    with pytest.raises(exception, match=match):
        validate_input(**kwargs)



import os.path
import shutil

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.tools import read_modelfit_results, run_tool
from pharmpy.workflows import ModelDatabase


def test_allometry(tmp_path, testdata):
    with chdir(tmp_path):
        for path in (testdata / 'nonmem' / 'models').glob('pheno5.*'):
            shutil.copy2(path, tmp_path)
        # shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        # shutil.copy2(testdata / 'nonmem' / 'sdtab1', tmp_path)

        model = Model.parse_model('pheno5.mod')
        results = read_modelfit_results('pheno5.mod')
        datainfo = model.datainfo.replace(path=tmp_path / 'pheno5.csv')
        model = model.replace(datainfo=datainfo)
        res = run_tool('allometry', model=model, results=results, allometric_variable='WGT')
        assert len(res.summary_models) == 2

        db: ModelDatabase = res.tool_database.model_database
        sep = os.path.sep
        model_name = 'scaled_model'
        assert str(db.retrieve_model(model_name).datainfo.path).endswith(
            f'{sep}allometry_dir1{sep}models{sep}.datasets{sep}input_model.csv'
        )
        path = db.retrieve_file(model_name, f'{model_name}.lst')
        with open(path, 'r') as fh:
            while line := fh.readline():
                # NOTE: Skip date, time, description etc
                if line[:6] == '$DATA ':
                    assert line == f'$DATA ..{sep}.datasets{sep}input_model.csv IGNORE=@\n'
                    break



from typing import Literal

import pytest

from pharmpy.internals.fn.type import _match, check_list


def test_check_list():
    with pytest.raises(ValueError):
        check_list("my", "a", ("b", "c", "d"))

    check_list("my", "a", ("a", "b", "c"))


def test_match():
    assert _match(dict[str, int], {'a': 1})
    assert not _match(dict[str, int], {'a': '1'})
    assert _match(dict[Literal['a', 'b'], int], {'b': 1})
    assert _match(dict[Literal['a', 'b'], str], {'b': 'c'})
    assert not _match(dict[Literal['a', 'b'], int], {'c': 1})
    assert _match(dict[int, int], {1: 2})
    assert not _match(dict[int, int], {1: '2'})



from pharmpy.deps import sympy

_smallz = 2.8e-103


class PHI(sympy.Function):
    pass


def PEXP(x):
    return sympy.Piecewise((sympy.exp(100), x > 100), (sympy.exp(x), True))


def PLOG(x):
    return sympy.Piecewise((sympy.log(_smallz), x < _smallz), (sympy.log(x), True))


def LOG10(x):
    return sympy.log(x, 10)


def PLOG10(x):
    return sympy.Piecewise((sympy.log(_smallz, 10), x < _smallz), (sympy.log(x, 10), True))


def PSQRT(x):
    return sympy.Piecewise((0, x < 0), (sympy.sqrt(x), True))


def INT(x):
    return sympy.sign(x) * sympy.floor(sympy.Abs(x))


def PDZ(x):
    return sympy.Piecewise((1 / _smallz, abs(x) < _smallz), (1 / x, True))


def PZR(x):
    return sympy.Piecewise((_smallz, abs(x) < _smallz), (x, True))


def PNP(x):
    return sympy.Piecewise((_smallz, x < _smallz), (x, True))


def PHE(x):
    return sympy.Piecewise((100, x > 100), (x, True))


def PNG(x):
    return sympy.Piecewise((0, x < 0), (x, True))



import pytest
import sympy

from pharmpy.model import Assignment
from pharmpy.modeling import (
    add_peripheral_compartment,
    create_basic_pk_model,
    set_mixed_mm_fo_elimination,
    set_tmdd,
)
from pharmpy.modeling.tmdd import _validate_dv_types


def S(x):
    return sympy.Symbol(x)


def test_full(pheno_path, load_model_for_test):
    model = load_model_for_test(pheno_path)
    model = set_tmdd(model, type="full")
    assert model.statements.ode_system.eqs[0].rhs == sympy.sympify(
        'A_CENTRAL(t)*(-A_TARGET(t)*KON/V - CL/V) + A_COMPLEX(t)*KOFF'
    )


def test_qss(pheno_path, load_model_for_test):
    model = load_model_for_test(pheno_path)
    model = set_tmdd(model, type="qss")
    assert model.statements.ode_system.eqs[0].rhs == sympy.sympify(
        '-A_CENTRAL(t)*CL*LAFREE/V - A_TARGET(t)*KINT*LAFREE/(KD + LAFREE) - CL*LAFREE/V'
    )


def test_qss_1c(pheno_path, load_model_for_test):
    model = load_model_for_test(pheno_path)
    model = add_peripheral_compartment(model)
    tmdd_model = set_tmdd(model, type="qss")
    assert (
        str(tmdd_model.statements.ode_system.eqs[0].rhs)
        == """-CL*LAFREE/V1 - KINT*LAFREE*A_TARGET(t)/(KD + LAFREE) - LAFREE*QP1/V1 + QP1*A_PERIPHERAL1(t)/VP1"""
    )
    assert (
        str(tmdd_model.statements.ode_system.eqs[1].rhs)
        == """LAFREE*QP1/V1 - QP1*A_PERIPHERAL1(t)/VP1"""
    )
    assert (
        str(tmdd_model.statements.ode_system.eqs[2].rhs)
        == """KSYN*V1 + (-KDEG - LAFREE*(-KDEG + KINT)/(KD + LAFREE))*A_TARGET(t)"""
    )

    model2 = add_peripheral_compartment(model)
    tmdd_model = set_tmdd(model2, type="qss")


def test_cr(pheno_path, load_model_for_test):
    model = load_model_for_test(pheno_path)
    model = set_tmdd(model, type="cr")
    assert model.statements.ode_system.eqs[1].rhs == sympy.sympify(
        '(-KINT - KOFF)*A_COMPLEX(t) + (KON*R_0 - KON*A_COMPLEX(t)/V)*A_CENTRAL(t)'
    )


def test_crib(pheno_path, load_model_for_test):
    model = load_model_for_test(pheno_path)
    model = set_tmdd(model, type="crib")
    assert model.statements.ode_system.eqs[1].rhs == sympy.sympify(
        '-KINT*A_COMPLEX(t) + (KON*R_0 - KON*A_COMPLEX(t)/V)*A_CENTRAL(t)'
    )


def test_ib(pheno_path, load_model_for_test):
    model = load_model_for_test(pheno_path)
    model = set_tmdd(model, type="ib")
    assert model.statements.ode_system.eqs[2].rhs == sympy.sympify(
        '-KDEG*A_TARGET(t) - KON*A_CENTRAL(t)*A_TARGET(t)/V + KSYN*V'
    )


def test_mmapp(pheno_path, load_model_for_test):
    model = load_model_for_test(pheno_path)
    model = add_peripheral_compartment(model)
    model = set_tmdd(model, type="mmapp")
    assert (
        str(model.statements.ode_system.eqs[0].rhs)
        == """QP1*A_PERIPHERAL1(t)/VP1 + (-CL/V1 - KINT*A_TARGET(t)/(KM + A_CENTRAL(t)/V1) - QP1/V1)*A_CENTRAL(t)"""
    )


def test_wagner(pheno_path, load_model_for_test):
    model = load_model_for_test(pheno_path)
    model = add_peripheral_compartment(model)
    model = set_tmdd(model, type="wagner")


def test_tmdd_with_mixed_fo_mm_elimination(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = add_peripheral_compartment(model)
    model = set_mixed_mm_fo_elimination(model)
    qss = set_tmdd(model, 'qss')
    assert qss


def test_wagner_1c(pheno_path, load_model_for_test):
    model = load_model_for_test(pheno_path)
    model = add_peripheral_compartment(model)
    wagner = set_tmdd(model, type="wagner")
    assert (
        str(wagner.statements.ode_system.eqs[0].rhs)
        == """-CL*LAFREE/V1 + KINT*LAFREE - KINT*A_CENTRAL(t) - LAFREE*QP1/V1 + QP1*A_PERIPHERAL1(t)/VP1"""
    )
    assert (
        str(wagner.statements.ode_system.eqs[1].rhs)
        == """LAFREE*QP1/V1 - QP1*A_PERIPHERAL1(t)/VP1"""
    )

    model = add_peripheral_compartment(model)
    wagner = set_tmdd(model, type="wagner")
    assert (
        str(wagner.statements.ode_system.eqs[1].rhs)
        == """LAFREE*QP1/V1 - QP1*A_PERIPHERAL1(t)/VP1"""
    )


@pytest.mark.parametrize(
    ('model_name', 'dv_types', 'expected'),
    (
        (
            'full',
            {'drug': 1, 'target': 2, 'complex': 3},
            {sympy.Symbol('Y'): 1, sympy.Symbol('Y_TARGET'): 2, sympy.Symbol('Y_COMPLEX'): 3},
        ),
        (
            'qss',
            {'drug': 1, 'target': 2, 'complex': 3},
            {sympy.Symbol('Y'): 1, sympy.Symbol('Y_TARGET'): 2, sympy.Symbol('Y_COMPLEX'): 3},
        ),
        (
            'ib',
            {'drug': 1, 'target': 2, 'complex': 3},
            {sympy.Symbol('Y'): 1, sympy.Symbol('Y_TARGET'): 2, sympy.Symbol('Y_COMPLEX'): 3},
        ),
        ('mmapp', {'drug': 1, 'target': 2}, {sympy.Symbol('Y'): 1, sympy.Symbol('Y_TARGET'): 2}),
        ('cr', {'drug': 1, 'complex': 2}, {sympy.Symbol('Y'): 1, sympy.Symbol('Y_COMPLEX'): 2}),
        ('wagner', {'drug': 1, 'complex': 2}, {sympy.Symbol('Y'): 1, sympy.Symbol('Y_COMPLEX'): 2}),
        ('crib', {'drug': 1, 'complex': 2}, {sympy.Symbol('Y'): 1, sympy.Symbol('Y_COMPLEX'): 2}),
        ('crib', {'drug': 1, 'complex': 3}, {sympy.Symbol('Y'): 1, sympy.Symbol('Y_COMPLEX'): 3}),
        (
            'full',
            {'target': 2, 'complex': 3},
            {sympy.Symbol('Y'): 1, sympy.Symbol('Y_TARGET'): 2, sympy.Symbol('Y_COMPLEX'): 3},
        ),
        (
            'cr',
            {'drug': 1, 'complex': 2, 'target': 3},
            {sympy.Symbol('Y'): 1, sympy.Symbol('Y_COMPLEX'): 2},
        ),
        (
            'mmapp',
            {'drug': 1, 'target': 2, 'complex': 3},
            {sympy.Symbol('Y'): 1, sympy.Symbol('Y_TARGET'): 2},
        ),
    ),
    ids=repr,
)
def test_full_multiple_dvs(pheno_path, load_model_for_test, model_name, dv_types, expected):
    model = load_model_for_test(pheno_path)
    model = set_tmdd(model, model_name, dv_types)
    assert model.dependent_variables == expected
    assert len(model.random_variables.epsilons) > 1


def test_multiple_dvs(load_model_for_test, pheno_path):
    central_amount = sympy.Function("A_CENTRAL")(S('t'))
    complex_amount = sympy.Function("A_COMPLEX")(S('t'))

    model = load_model_for_test(pheno_path)
    model1 = set_tmdd(model, 'qss', {'complex': 2, 'target_tot': 3})
    ass1 = Assignment(S("F"), S("LAFREEF") / S("V"))
    assert model1.statements.find_assignment("F") == ass1
    assert S("Y_COMPLEX") in model1.statements.free_symbols
    assert S("Y_TOTTARGET") in model1.statements.free_symbols

    model2 = set_tmdd(model, 'qss', {'drug_tot': 1, 'complex': 2, 'target_tot': 3})
    ass2 = Assignment(S("F"), central_amount / S("V"))
    assert model2.statements.find_assignment("F") == ass2
    assert S("Y_COMPLEX") in model2.statements.free_symbols
    assert S("Y_TOTTARGET") in model2.statements.free_symbols

    model1 = set_tmdd(model, 'full', {'complex': 2, 'target_tot': 3})
    ass1 = Assignment(S("F"), central_amount / S("V"))
    assert model1.statements.find_assignment("F") == ass1
    assert S("Y_COMPLEX") in model1.statements.free_symbols
    assert S("Y_TOTTARGET") in model1.statements.free_symbols

    model2 = set_tmdd(model, 'full', {'drug_tot': 1, 'complex': 2, 'target_tot': 3})
    ass2 = Assignment(S("F"), (central_amount + complex_amount) / S("V"))
    assert model2.statements.find_assignment("F") == ass2
    assert S("Y_COMPLEX") in model2.statements.free_symbols
    assert S("Y_TOTTARGET") in model2.statements.free_symbols

    model1 = set_tmdd(model, 'cr', {'complex': 2})
    ass1 = Assignment(S("F"), central_amount / S("V"))
    assert model1.statements.find_assignment("F") == ass1
    assert S("Y_COMPLEX") in model1.statements.free_symbols

    model2 = set_tmdd(model, 'cr', {'drug_tot': 1, 'complex': 2})
    ass2 = Assignment(S("F"), (central_amount + complex_amount) / S("V"))
    assert model2.statements.find_assignment("F") == ass2
    assert S("Y_COMPLEX") in model2.statements.free_symbols

    model1 = set_tmdd(model, 'wagner', {'drug_tot': 1, 'complex': 2})
    ass1 = Assignment(S("F"), central_amount / S("V"))
    assert model1.statements.find_assignment("F") == ass1
    assert S("Y_COMPLEX") in model1.statements.free_symbols

    model1 = set_tmdd(model, 'mmapp', {'drug': 1, 'target_tot': 3})
    ass1 = Assignment(S("F"), central_amount / S("V"))
    assert model1.statements.find_assignment("F") == ass1
    assert S("Y_TOTTARGET") in model1.statements.free_symbols

    model = create_basic_pk_model('iv')
    model1 = set_tmdd(model, 'qss', {'complex': 2, 'target_tot': 3})
    ass1 = Assignment(S("IPRED"), S("LAFREEF") / S("VC"))
    assert model1.statements.find_assignment("IPRED") == ass1
    assert S("Y_COMPLEX") in model1.statements.free_symbols
    assert S("Y_TOTTARGET") in model1.statements.free_symbols

    model2 = set_tmdd(model, 'qss', {'drug_tot': 1, 'complex': 2, 'target_tot': 3})
    ass2 = Assignment(S("IPRED"), central_amount / S("VC"))
    assert model2.statements.find_assignment("IPRED") == ass2
    assert S("Y_COMPLEX") in model2.statements.free_symbols
    assert S("Y_TOTTARGET") in model2.statements.free_symbols


def test_validation(load_model_for_test, testdata):
    _validate_dv_types(dv_types={'drug_tot': 1, 'target_tot': 2, 'complex': 3})
    _validate_dv_types(dv_types={'drug': 1, 'target_tot': 2, 'complex': 3})
    _validate_dv_types(dv_types={'drug': 1, 'target': 2, 'complex': 3})

    with pytest.raises(AssertionError):
        _validate_dv_types(dv_types={'drug': 1, 'target': 1, 'complex': 2})
    with pytest.raises(
        ValueError, match='Only drug can have DVID = 1. Please choose another DVID.'
    ):
        _validate_dv_types(dv_types={'target': 1, 'complex': 2})
    with pytest.raises(
        ValueError,
        match='Invalid dv_types key "taget". Allowed keys are: "drug", "target", "complex", '
        '"drug_tot" and "target_tot".',
    ):
        _validate_dv_types(dv_types={'taget': 1})



from functools import partial
from typing import List, Literal, Optional

from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.deps.scipy import stats
from pharmpy.internals.fn.signature import with_same_arguments_as
from pharmpy.internals.fn.type import with_runtime_arguments_type_check
from pharmpy.model import (
    Assignment,
    EstimationStep,
    EstimationSteps,
    Model,
    NormalDistribution,
    Parameter,
    Parameters,
    RandomVariables,
    Statements,
)
from pharmpy.modeling import (
    add_population_parameter,
    add_time_after_dose,
    create_symbol,
    get_mdv,
    has_proportional_error_model,
    set_combined_error_model,
    set_iiv_on_ruv,
    set_initial_estimates,
    set_power_on_ruv,
    set_proportional_error_model,
)
from pharmpy.modeling.blq import has_blq_transformation
from pharmpy.modeling.error import remove_error_model, set_time_varying_error_model
from pharmpy.tools import (
    summarize_errors,
    summarize_individuals,
    summarize_individuals_count_table,
    summarize_modelfit_results,
)
from pharmpy.tools.common import summarize_tool, update_initial_estimates
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder, call_workflow
from pharmpy.workflows.results import ModelfitResults

from .results import RUVSearchResults, calculate_results

SKIP = frozenset(('IIV_on_RUV', 'power', 'combined', 'time_varying'))


def create_workflow(
    model: Optional[Model] = None,
    results: Optional[ModelfitResults] = None,
    groups: int = 4,
    p_value: float = 0.001,
    skip: Optional[List[Literal[tuple(SKIP)]]] = None,
    max_iter: int = 3,
    dv: Optional[int] = None,
    strictness: Optional[str] = "minimization_successful or (rounding_errors and sigdigs>=0.1)",
):
    """Run the ruvsearch tool. For more details, see :ref:`ruvsearch`.

    Parameters
    ----------
    model : Model
        Pharmpy model
    results : ModelfitResults
        Results of model
    groups : int
        The number of bins to use for the time varying models
    p_value : float
        The p-value to use for the likelihood ratio test
    skip : list of {'IIV_on_RUV', 'power', 'combined', 'time_varying'}
        A list of models to not attempt.
    max_iter :  int
        Number of iterations to run (1, 2, or 3). For models with BLQ only one iteration is supported.
    dv : int
        Which DV to assess the error model for.
    strictness : str or None
        Strictness criteria

    Returns
    -------
    RUVSearchResults
        Ruvsearch tool result object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model
    >>> from pharmpy.tools import run_ruvsearch, load_example_modelfit_results
    >>> model = load_example_model("pheno")
    >>> results = load_example_modelfit_results("pheno")
    >>> run_ruvsearch(model=model, results=results)      # doctest: +SKIP

    """

    wb = WorkflowBuilder(name="ruvsearch")
    start_task = Task(
        'start_ruvsearch', start, model, results, groups, p_value, skip, max_iter, dv, strictness
    )
    wb.add_task(start_task)
    task_results = Task('results', _results)
    wb.add_task(task_results, predecessors=[start_task])
    return Workflow(wb)


def create_iteration_workflow(model_entry, groups, cutoff, skip, current_iteration, dv):
    wb = WorkflowBuilder()

    start_task = Task('start_iteration', _start_iteration, model_entry)
    wb.add_task(start_task)

    task_base_model = Task(
        'create_base_model', partial(_create_base_model, current_iteration=current_iteration, dv=dv)
    )
    wb.add_task(task_base_model, predecessors=start_task)

    tasks = []
    if 'IIV_on_RUV' not in skip:
        task_iiv = Task(
            'create_iiv_on_ruv_model',
            partial(_create_iiv_on_ruv_model, current_iteration=current_iteration, dv=None),
        )
        tasks.append(task_iiv)
        wb.add_task(task_iiv, predecessors=task_base_model)

    if 'power' not in skip and 'combined' not in skip:
        task_power = Task(
            'create_power_model',
            partial(_create_power_model, current_iteration=current_iteration, dv=None),
        )
        wb.add_task(task_power, predecessors=task_base_model)
        tasks.append(task_power)
        task_combined = Task(
            'create_combined_error_model',
            partial(_create_combined_model, current_iteration=current_iteration),
        )
        wb.add_task(task_combined, predecessors=task_base_model)
        tasks.append(task_combined)

    if 'time_varying' not in skip:
        for i in range(1, groups):
            tvar = partial(
                _create_time_varying_model,
                groups=groups,
                i=i,
                current_iteration=current_iteration,
                dv=None,
            )
            task = Task(f"create_time_varying_model{i}", tvar)
            tasks.append(task)
            wb.add_task(task, predecessors=task_base_model)

    fit_wf = create_fit_workflow(n=1 + len(tasks))
    wb.insert_workflow(fit_wf, predecessors=[task_base_model] + tasks)
    post_pro = partial(post_process, cutoff=cutoff, current_iteration=current_iteration, dv=dv)
    task_post_process = Task('post_process', post_pro)
    wb.add_task(task_post_process, predecessors=[start_task] + fit_wf.output_tasks)

    return Workflow(wb)


def proportional_error_workflow(model_entry):
    wb = WorkflowBuilder()

    prop_start = Task('Check_proportional', _start_iteration, model_entry)
    wb.add_task(prop_start)

    if not has_proportional_error_model(model_entry.model):
        convert_to_prop_task = Task("convert_to_proportional", _change_proportional_model)
        wb.add_task(convert_to_prop_task, predecessors=prop_start)

        fit_wf = create_fit_workflow(n=1)
        wb.insert_workflow(fit_wf, predecessors=convert_to_prop_task)
    return Workflow(wb)


def _change_proportional_model(model_entry):
    model = model_entry.model
    model = model.replace(
        name='prop_error',
        description='Input model with proportional error model',
    )
    model = set_proportional_error_model(model)
    return ModelEntry.create(model, modelfit_results=None)


def start(context, input_model, input_res, groups, p_value, skip, max_iter, dv, strictness):
    cutoff = float(stats.chi2.isf(q=p_value, df=1))
    if skip is None:
        skip = []

    input_model_entry = ModelEntry.create(input_model, modelfit_results=input_res)
    # Check if model has a proportional error
    proportional_workflow = proportional_error_workflow(input_model_entry)
    model_entry = call_workflow(proportional_workflow, 'Convert_error_model', context)

    if model_entry.model == input_model_entry.model:
        selected_model_entries = [model_entry]
    else:
        selected_model_entries = [input_model_entry, model_entry]
    cwres_models = []
    tool_database = None
    for current_iteration in range(1, max_iter + 1):
        wf = create_iteration_workflow(model_entry, groups, cutoff, skip, current_iteration, dv=dv)
        res, best_model_entry, selected_model_name = call_workflow(
            wf, f'results{current_iteration}', context
        )
        cwres_models.append(res.cwres_models)
        tool_database = res.tool_database

        if not selected_model_name.startswith('base'):
            selected_model_entries.append(best_model_entry)

        model_entry = best_model_entry

        if selected_model_name.startswith('base'):
            break
        elif selected_model_name.startswith('time_varying'):
            skip.append('time_varying')
        else:
            skip.append(selected_model_name)

    # Check that there actually occured an improvement from the initial model.
    delta_ofv = input_model_entry.modelfit_results.ofv - model_entry.modelfit_results.ofv
    if delta_ofv < cutoff:
        model_entry = input_model_entry

    selected_models = [model_entry.model for model_entry in selected_model_entries]
    model_results = [model_entry.modelfit_results for model_entry in selected_model_entries]

    sumind = summarize_individuals(selected_models, model_results)
    sumcount = summarize_individuals_count_table(df=sumind)
    sum_models = summarize_modelfit_results(model_results)
    sum_models['step'] = list(range(len(sum_models)))
    summf = sum_models.reset_index().set_index(['step', 'model'])
    summary_tool = _create_summary_tool(selected_model_entries, cutoff, strictness)
    summary_errors = summarize_errors(m.modelfit_results for m in selected_model_entries)

    res = RUVSearchResults(
        cwres_models=pd.concat(cwres_models),
        summary_individuals=sumind,
        summary_individuals_count=sumcount,
        final_model=model_entry.model,
        summary_models=summf,
        summary_tool=summary_tool,
        summary_errors=summary_errors,
        tool_database=tool_database,
    )
    return res


def _create_summary_tool(selected_model_entries, cutoff, strictness):
    selected_models = [model_entry.model for model_entry in selected_model_entries]
    model_names = [model.name for model in selected_models]
    iteration_map = {model.name: model_names.index(model.name) for model in selected_models}

    base_model_entry = selected_model_entries[0]
    ruvsearch_model_entries = selected_model_entries[1:]

    sum_tool = summarize_tool(
        ruvsearch_model_entries, base_model_entry, 'ofv', cutoff, strictness
    ).reset_index()
    sum_tool['step'] = sum_tool['model'].map(iteration_map)
    sum_tool_by_iter = sum_tool.set_index(['step', 'model']).sort_index()

    # FIXME: Workaround since rank_models will exclude ranking of base model since dofv will be 0
    sum_tool_by_iter.loc[
        (0, base_model_entry.model.name), 'ofv'
    ] = base_model_entry.modelfit_results.ofv
    sum_tool_by_iter.loc[(0, base_model_entry.model.name), 'dofv'] = 0

    return sum_tool_by_iter.drop(columns=['rank'])


def _start_iteration(model_or_model_entry):
    return model_or_model_entry


def _results(res):
    return res


def post_process(context, start_model_entry, *model_entries, cutoff, current_iteration, dv):
    res = calculate_results(model_entries)
    best_model_unfitted, selected_model_name = _create_best_model(
        start_model_entry, res, current_iteration, cutoff=cutoff, dv=dv
    )
    if best_model_unfitted is not None:
        fit_wf = create_fit_workflow(modelentries=[best_model_unfitted])
        best_model_entry = call_workflow(fit_wf, f'fit{current_iteration}', context)
        if best_model_entry.modelfit_results is not None:
            best_model_check = [
                best_model_entry.modelfit_results.ofv,
                best_model_entry.modelfit_results.residuals,
                best_model_entry.modelfit_results.predictions,
            ]
            if all(check is not None for check in best_model_check):
                delta_ofv = (
                    start_model_entry.modelfit_results.ofv - best_model_entry.modelfit_results.ofv
                )
                if delta_ofv > cutoff:
                    return (res, best_model_entry, selected_model_name)

    return (res, start_model_entry, f"base_{current_iteration}")


def _create_base_model(input_model_entry, current_iteration, dv):
    input_model = input_model_entry.model
    theta = Parameter('theta', 0.1)
    omega = Parameter('omega', 0.01, lower=0)
    sigma = Parameter('sigma', 1, lower=0)
    params = Parameters((theta, omega, sigma))

    eta_name = 'eta_base'
    eta = NormalDistribution.create(eta_name, 'iiv', 0, omega.symbol)
    sigma_name = 'epsilon'
    sigma = NormalDistribution.create(sigma_name, 'ruv', 0, sigma.symbol)
    rvs = RandomVariables.create([eta, sigma])

    y = Assignment(
        sympy.Symbol('Y'), theta.symbol + sympy.Symbol(eta_name) + sympy.Symbol(sigma_name)
    )
    statements = Statements([y])

    name = f'base_{current_iteration}'

    est = EstimationStep.create('foce', interaction=True, maximum_evaluations=9999)

    base_model = Model.create(
        parameters=params,
        random_variables=rvs,
        statements=statements,
        name=name,
        description=name,
        estimation_steps=EstimationSteps.create([est]),
        dependent_variables={y.symbol: 1},
    )
    base_model = base_model.replace(dataset=_create_dataset(input_model_entry, dv))
    return ModelEntry.create(base_model, modelfit_results=None, parent=input_model)


def _create_iiv_on_ruv_model(input_model_entry, current_iteration, dv):
    input_model = input_model_entry.model
    model = set_iiv_on_ruv(input_model, dv)
    name = f'IIV_on_RUV_{current_iteration}'
    model = model.replace(name=name, description=name)
    return ModelEntry.create(model, modelfit_results=None, parent=input_model)


def _create_power_model(input_model_entry, current_iteration, dv):
    input_model = input_model_entry.model
    model = set_power_on_ruv(
        input_model, ipred='IPRED', lower_limit=None, zero_protection=True, dv=dv
    )
    name = f'power_{current_iteration}'
    model = model.replace(name=name, description=name)
    return ModelEntry.create(model, modelfit_results=None, parent=input_model)


def _create_time_varying_model(input_model_entry, groups, i, current_iteration, dv):
    input_model = input_model_entry.model
    quantile = i / groups
    cutoff = input_model.dataset['TAD'].quantile(q=quantile)
    model = set_time_varying_error_model(input_model, cutoff=cutoff, idv='TAD', dv=dv)
    name = f"time_varying{i}_{current_iteration}"
    model = model.replace(name=name, description=name)
    return ModelEntry.create(model, modelfit_results=None, parent=input_model)


def _create_combined_model(input_model_entry, current_iteration):
    input_model = input_model_entry.model
    model = remove_error_model(input_model)
    sset = model.statements
    ruv_prop = create_symbol(model, 'epsilon_p')
    ruv_add = create_symbol(model, 'epsilon_a')
    ipred = sympy.Symbol('IPRED')
    s = sset[0]
    assert isinstance(s, Assignment)
    s = Assignment(s.symbol, s.expression + ruv_prop + ruv_add / ipred)

    prop_name = 'sigma_prop'
    model = add_population_parameter(model, prop_name, 1, lower=0)
    df = model.dataset
    assert df is not None
    df = df.copy()
    df['IPRED'].replace(0, 2.225e-307, inplace=True)
    model = model.replace(dataset=df)
    ipred_min = df['IPRED'].min()
    sigma_add_init = abs(ipred_min) / 2 if ipred_min != 0 else 0.001
    add_name = 'sigma_add'
    model = add_population_parameter(model, add_name, sigma_add_init, lower=0)

    eps_prop = NormalDistribution.create(ruv_prop.name, 'ruv', 0, sympy.Symbol(prop_name))
    eps_add = NormalDistribution.create(ruv_add.name, 'ruv', 0, sympy.Symbol(add_name))
    name = f'combined_{current_iteration}'
    model = model.replace(
        statements=s + sset[1:],
        random_variables=model.random_variables + [eps_prop, eps_add],
        name=name,
        description=name,
    )
    return ModelEntry.create(model, modelfit_results=None, parent=input_model)


def _create_dataset(input_model_entry: ModelEntry, dv):
    # Non-observations have already been filtered
    input_model, results = input_model_entry.model, input_model_entry.modelfit_results
    assert results is not None
    residuals = results.residuals
    assert residuals is not None
    input_dataset = input_model.dataset
    assert input_dataset is not None
    if dv is not None:
        observation_label = input_model.datainfo.dv_column.name
        input_dataset = input_dataset.query(f'{observation_label} != 0').reset_index(
            drop=True
        )  # filter non-observations
        indices = input_dataset.index[input_dataset['DVID'] == dv].tolist()
        residuals = residuals.iloc[indices]
    cwres = residuals['CWRES'].reset_index(drop=True)
    if has_blq_transformation(input_model):
        cwres = cwres.loc[cwres != 0]
    predictions = results.predictions
    assert predictions is not None
    if 'CIPREDI' in predictions:
        ipredcol = 'CIPREDI'
    elif 'IPRED' in predictions:
        ipredcol = 'IPRED'
    else:
        raise ValueError("Need CIPREDI or IPRED")
    ipred = predictions[ipredcol].reset_index(drop=True)
    mdv = get_mdv(input_model)
    mdv = mdv.reset_index(drop=True)
    label_id = input_model.datainfo.id_column.name
    input_id = input_dataset[label_id].astype('int64').squeeze().reset_index(drop=True)
    input_model = add_time_after_dose(input_model)
    tad_label = input_model.datainfo.descriptorix['time after dose'][0].name
    tad = input_model.dataset[tad_label].squeeze().reset_index(drop=True)
    df = pd.concat([mdv, input_id, tad, ipred], axis=1)
    df = df[df['MDV'] == 0].reset_index(drop=True)
    df = pd.concat([df, cwres], axis=1).rename(columns={'CWRES': 'DV', ipredcol: 'IPRED'})
    df = df.loc[df['DV'].notna()]
    return df


def _time_after_dose(model):
    if 'TAD' in model.dataset:
        pass
    else:
        model = add_time_after_dose(model)
    return model


def _create_best_model(model_entry, res, current_iteration, dv, groups=4, cutoff=3.84):
    if not res.cwres_models.empty and any(res.cwres_models['dofv'] > cutoff):
        model = update_initial_estimates(model_entry.model, model_entry.modelfit_results)
        selected_model_name = f'base_{current_iteration}'
        idx = res.cwres_models['dofv'].idxmax()
        name = idx[0]

        if current_iteration == 1:
            base_description = ''
        else:
            base_description = model.description + '+'
        model = model.replace(
            name=f'best_ruvsearch_{current_iteration}', description=base_description + name
        )

        if name.startswith('power'):
            model = set_power_on_ruv(model, dv=dv)
            model = set_initial_estimates(
                model,
                {
                    'power1': res.cwres_models['parameters']
                    .loc['power', 1, current_iteration]
                    .get('theta')
                    + 1
                },
            )
        elif name.startswith('IIV_on_RUV'):
            model = set_iiv_on_ruv(model, dv=dv)
            model = set_initial_estimates(
                model,
                {
                    'IIV_RUV1': res.cwres_models['parameters']
                    .loc['IIV_on_RUV', 1, current_iteration]
                    .get('omega')
                },
            )
        elif name.startswith('time_varying'):
            model = _time_after_dose(model)
            i = int(name[-1])
            quantile = i / groups
            df = _create_dataset(model_entry, dv=dv)
            tad = df['TAD']
            cutoff_tvar = tad.quantile(q=quantile)
            model = set_time_varying_error_model(model, cutoff=cutoff_tvar, idv='TAD', dv=dv)
            model = set_initial_estimates(
                model,
                {
                    'time_varying': res.cwres_models['parameters']
                    .loc[f"time_varying{i}", 1, current_iteration]
                    .get('theta')
                },
            )
        else:
            model = set_combined_error_model(model, dv=dv)
            model = set_initial_estimates(
                model,
                {
                    'sigma_prop': res.cwres_models['parameters']
                    .loc['combined', 1, current_iteration]
                    .get('sigma_prop'),
                    'sigma_add': res.cwres_models['parameters']
                    .loc['combined', 1, current_iteration]
                    .get('sigma_add'),
                },
            )

        best_model_entry = ModelEntry.create(model, modelfit_results=None, parent=model_entry.model)
        selected_model_name = name
    else:
        best_model_entry = None
        selected_model_name = None

    return best_model_entry, selected_model_name


@with_runtime_arguments_type_check
@with_same_arguments_as(create_workflow)
def validate_input(model, results, groups, p_value, skip, max_iter, dv, strictness):
    if groups <= 0:
        raise ValueError(f'Invalid `groups`: got `{groups}`, must be >= 1.')

    if not 0 < p_value <= 1:
        raise ValueError(f'Invalid `p_value`: got `{p_value}`, must be a float in range (0, 1].')

    if max_iter < 1 or max_iter > 3:
        raise ValueError(f'Invalid `max_iter`: got `{max_iter}`, must be int in range [1, 3].')

    if model is not None:
        if results is None:
            raise ValueError('Invalid `results`: modelfit results must be provided.')

        residuals = results.residuals
        if residuals is None or 'CWRES' not in residuals:
            raise ValueError(
                f'Invalid `results`: please check {model.name}.mod file to'
                f' make sure ID, TIME, CWRES are in $TABLE.'
            )

        predictions = results.predictions
        if predictions is None or ('CIPREDI' not in predictions and 'IPRED' not in predictions):
            raise ValueError(
                f'Invalid `results`: please check {model.name}.mod file to'
                f' make sure ID, TIME, CIPREDI (or IPRED) are in $TABLE.'
            )

        if has_blq_transformation(model) and max_iter > 1:
            raise ValueError(
                f'Invalid `max_iter`: got `{max_iter}`,only 1 iteration is supported '
                f'for models with BLQ transformation.'
            )

        if dv:
            if 'DVID' not in model.dataset.columns:
                raise ValueError("No column DVID in dataset.")
            else:
                if dv not in set(model.dataset['DVID']):
                    raise ValueError(f"No DVID = {dv} in dataset.")

    if strictness is not None and "rse" in strictness.lower():
        if model.estimation_steps[-1].parameter_uncertainty_method is None:
            raise ValueError(
                'parameter_uncertainty_method not set for model, cannot calculate relative standard errors.'
            )



import itertools

from pharmpy.deps import numpy as np
from pharmpy.internals.math import corr2cov, nearest_positive_semidefinite
from pharmpy.modeling import fix_or_unfix_parameters, set_initial_estimates


def calculate_parcov_inits(model, ie, ncovs):
    """Get dict of new updated inits for parcov block

    model already has a FREM block
    Initial estimates for the parcov block is calculated given correlations of individual etas
    """
    dist = model.random_variables.iiv[-1]
    rvs = list(dist.names)
    eta_corr = ie[rvs].corr()
    eta_corr.fillna(value=1.0, inplace=True)  # Identical etas will get NaN as both diag and corr

    sigma = dist.variance
    inits = sigma.subs(model.parameters.inits)
    inits = np.array(inits).astype(np.float64)
    sd = np.sqrt(inits.diagonal())
    npars = len(sd) - ncovs

    cov = corr2cov(eta_corr.to_numpy(), sd)
    cov[cov == 0] = 0.0001
    cov = nearest_positive_semidefinite(cov)

    parcov_inits = cov[npars:, :npars]
    parcov_symb = sigma[npars:, :npars]

    param_inits = {
        parcov_symb[i, j].name: parcov_inits[i, j]
        for i, j in itertools.product(range(ncovs), range(npars))
    }
    return param_inits


def create_model3b(model1b, model3, model3_res, ncovs):
    """Create model 3b from model 3

    * Update parcov omega block
    * Set FIX pattern back from model1b
    * Use initial etas from model3
    """
    model3b = model3.replace(name='model_3b')
    ie = model3_res.individual_estimates
    parcov_inits = calculate_parcov_inits(model3, ie, ncovs)
    model3b = set_initial_estimates(model3b, parcov_inits)
    model3b = set_initial_estimates(
        model3b, model3b.random_variables.nearest_valid_parameters(model3b.parameters.inits)
    )
    model3b = fix_or_unfix_parameters(model3b, model1b.parameters.fix)

    model3b = model3b.replace(initial_individual_estimates=ie)
    return model3b



"""
Generic parser using lark.

Subclass :class:`GenericParser` (remember to set :attr:`GenericParser.grammar`
to point to your grammar file) to define a powerful parser.
"""
from __future__ import annotations

import re
from abc import ABC
from itertools import chain
from typing import Callable, Iterable, List, Optional, Sequence, Tuple, Union

from lark import Lark, Transformer, Tree, Visitor
from lark.lexer import Token

from . import prettyprint
from .tree import Leaf as ImmutableLeaf
from .tree import Tree as ImmutableTree

TYPES_OF_NEWLINE = frozenset(('CONT', 'NEWLINE'))
TYPES_OF_COMMENT = frozenset(('COMMENT',))


class AttrToken(ImmutableLeaf):
    def __repr__(self):
        return '%s(%s, %s)' % (self.__class__.__name__, repr(self.rule), repr(self.value))

    def __str__(self):
        return str(self.value)


def _newline_node():
    return AttrToken('NEWLINE', '\n')


def eval_token(token: AttrToken) -> Union[int, float, str]:
    """Evaluated value (str, int, float)."""
    if token.rule in ('DIGIT', 'INT', 'SIGNED_INT'):
        return int(token.value)
    elif token.rule in (
        'DECIMAL',
        'EXP',
        'FLOAT',
        'FLOAT2',
        'NUMBER',
        'NUMERIC',
        'SIGNED_FLOAT',
        'SIGNED_NUMBER',
    ):
        return float(token.value)
    elif token.rule == 'NEG_INF':
        return float('-INF')
    elif token.rule == 'POS_INF':
        return float('INF')
    else:
        return token.value


def _parse_create_input(rule: Optional[str], items):
    try:
        names, items = zip(*items.items())
    except AttributeError:
        if isinstance(items, str):
            raise TypeError(str)
        length = len(items)
        names: Tuple[Optional[str], ...] = (None,) * length
    if not items:
        raise ValueError(f'refusing empty tree {repr(rule)} (only tokens are childless)')

    return (rule, list(zip(names, items)), [])


class AttrTree(ImmutableTree['AttrTree', 'AttrToken']):
    """Tree with attribute access.

    Created in :meth:`GenericParser.parse` by :meth:`transform`, from :class:`lark.Tree`.

    Attributes:
        self.rule: Name, in common with :class:`.AttrToken`.
        self.rules: Names of children.
        self.eval: Transformed data type, in common with :class:`.AttrToken`.
        self.tokens: Recursive tokens as (flattened) list.
        self.debug: Treeview str, formatted for debugging.

    Can be instantiated with :meth:`.__init__`, via :class:`lark.Tree`, or alternative constructors:
        1. :meth:`.transform` (transform recursively object of class :class:`lark.Tree`).
        2. :meth:`.create` (create from nested iterators).
    """

    @staticmethod
    def create(rule: Optional[str], items) -> AttrTree:
        """Alternative constructor: Creates new tree from (possibly nested) iterables.

        Only non-iterable items become leaves (i.e. content of token nodes). All others are trees.

        Handling of missing names (e.g. lists):
            1. Tree: Children are moved to parent.
            2. Token: Native naming scheme, __ANON_%d, of Lark is used.

        Args:
            rule: Name of root tree (__ANON_0 if False).
            items: Child (tree) nodes or content (token) nodes.

        Raises:
            TypeError: 'items' not iterable or instance of 'str' (only leaves shall contain 'str').
            ValueError: 'items' empty (trees can't be empty).

        .. note:: Please follow convention of all lower/upper case for trees/tokens.
        """

        _anon_count = 0
        stack = [_parse_create_input(rule, items)]

        while True:
            rule_name, args, children = stack[-1]
            i = len(children)
            if i == len(args):
                # NOTE: We flatten anonymous trees
                flattened = tuple(
                    chain.from_iterable(
                        child if isinstance(child, tuple) else (child,) for child in children
                    )
                )
                tree = (
                    flattened
                    if not rule_name
                    else AttrTree(
                        rule_name,
                        flattened,
                    )
                )
                stack.pop()
                if stack:
                    # NOTE: We had this tree as child of parent
                    stack[-1][-1].append(tree)
                    continue
                else:
                    return tree if isinstance(tree, AttrTree) else AttrTree('__ANON_0', tree)

            name, thing = args[i]

            if isinstance(thing, (AttrTree, AttrToken)):
                # NOTE: Do not convert existing nodes
                children.append(thing)
            elif isinstance(thing, (str, int, float, type(None))):
                # NOTE: leaf
                if not name:
                    _anon_count += 1
                    name = '__ANON_%d' % (_anon_count,)
                children.append(AttrToken(name, str(thing)))
            else:
                # NOTE: recurse
                stack.append(_parse_create_input(name, thing))

    # -- public interface ----------------------------------------------
    @property
    def rules(self) -> List[str]:
        """All rules of (immediate) children."""
        return [node.rule for node in self.children]

    def first_index(self, rule: str) -> int:
        """Returns index of first child matching 'rule', or -1."""
        return next((i for i, child in enumerate(self.children) if child.rule == rule), -1)

    def find(self, rule) -> Optional[Union[AttrTree, AttrToken]]:
        """Returns first child matching 'rule', or None."""
        i = self.first_index(rule)
        return None if i == -1 else self.children[i]

    def first_branch(self, *rules) -> Optional[Union[AttrTree, AttrToken]]:
        current = self
        for rule in rules:
            if not isinstance(current, AttrTree):
                return None
            current = current.find(rule)
        return current

    def set(self, rule, new_child) -> AttrTree:
        """Sets first child matching rule. Raises if none."""
        i = self.first_index(rule)
        if i == -1:
            raise NoSuchRuleException(rule, self)

        return AttrTree(self.rule, self.children[:i] + (new_child,) + self.children[i + 1 :])

    def partition(
        self, rule
    ) -> Tuple[
        Tuple[Union[AttrTree, AttrToken], ...],
        Tuple[Union[AttrTree, AttrToken], ...],
        Tuple[Union[AttrTree, AttrToken], ...],
    ]:
        """Partition children into (head, item, tail).

        Search for child item 'rule' and return the part before it (head), the item, and the part
        after it (tail). If 'rule' is not found, return (children, [], [])."""

        i = self.first_index(rule)
        return (
            (self.children, (), ())
            if i == -1
            else (self.children[:i], (self.children[i],), self.children[i + 1 :])
        )

    def tree_walk(self):
        """Generator for iterating depth-first (i.e. parse order) over children."""
        for child in self.children:
            yield child
            if isinstance(child, AttrTree):
                yield from child.tree_walk()

    def remove(self, rule):
        """Remove all children with rule. Not recursively"""
        return AttrTree(
            self.rule,
            tuple(child for child in self.children if child.rule != rule),
        )

    def _clean_ws(self, new_children: Sequence[Union[AttrTree, AttrToken]]):
        new_children_clean = []
        last_index = len(new_children) - 1

        prev_rule = None

        for i, child in enumerate(new_children):
            if child.rule in TYPES_OF_NEWLINE:
                if prev_rule in TYPES_OF_NEWLINE or prev_rule == 'verbatim':
                    continue
                if i == last_index:
                    continue
            if re.search('\n{2,}', str(child)):
                new_children_clean.append(_newline_node())
            else:
                new_children_clean.append(child)

            prev_rule = child.rule

        return new_children_clean

    def add_node(self, node, following_node=None, comment=False):
        new_children = self._clean_ws(self.children) if comment else list(self.children)

        if following_node is None:
            if not comment:
                new_children.append(_newline_node())
            new_children.append(node)
        else:
            index = self.children.index(following_node)
            new_children.insert(index, node)

            newline_node = new_children[index - 1]
            new_children.insert(index + 1, newline_node)

        new_children_clean = self._clean_ws(new_children)
        return AttrTree(self.rule, tuple(new_children_clean))

    def add_comment_node(self, comment, adjacent_node=None):
        comment_node = AttrToken('COMMENT', f' ; {comment}')
        return self.add_node(comment_node, adjacent_node, comment=True)

    def add_newline_node(self):
        return AttrTree(self.rule, self.children + (_newline_node(),))

    def subtrees(self, rule) -> Iterable[AttrTree]:
        for child in self.children:
            if isinstance(child, AttrTree) and child.rule == rule:
                yield child

    def subtree(self, rule) -> AttrTree:
        try:
            return next(iter(self.subtrees(rule)))
        except StopIteration:
            raise NoSuchRuleException(f'No subtree "{rule}" in {repr(self)}.')

    def subtree_at(self, i: int) -> AttrTree:
        child = self.children[i]
        if isinstance(child, AttrTree):
            return child

        raise ValueError(f'The is no subtree at index {i}')

    def leaves(self, rule) -> Iterable[AttrToken]:
        for child in self.children:
            if isinstance(child, AttrToken) and child.rule == rule:
                yield child

    def leaf(self, rule) -> AttrToken:
        try:
            return next(iter(self.leaves(rule)))
        except StopIteration:
            raise NoSuchRuleException(f'No leaf "{rule}" in {repr(self)}.')

    def leaf_at(self, i: int) -> AttrToken:
        child = self.children[i]
        if isinstance(child, AttrToken):
            return child

        raise ValueError(f'The is no subtree at index {i}')

    def replace_first(self, child: Union[AttrTree, AttrToken]) -> AttrTree:
        i = self.first_index(child.rule)
        if i == -1 or self.children[i] == child:
            return self

        new_children = self.children[:i] + (child,) + self.children[i + 1 :]
        return AttrTree(self.rule, new_children)

    def map(
        self, fn: Callable[[Union[AttrTree, AttrToken]], Union[AttrTree, AttrToken]]
    ) -> AttrTree:
        return AttrTree(self.rule, tuple(map(fn, self.children)))

    @property
    def tokens(self) -> Iterable[AttrToken]:
        """All tokens in depth-first order."""
        for child in self.tree_walk():
            if isinstance(child, AttrToken):
                yield child

    @property
    def debug(self, *args, **kwargs):
        """Debug formatted tree structure."""
        return str(prettyprint.transform(self, *args, **kwargs))

    def treeprint(self, indent=''):
        """Prints debug formatted tree structure."""
        print(self.debug)

    # -- private methods -----------------------------------------------
    def __len__(self):
        return len(self.children)

    def __str__(self):
        return ''.join(str(x) for x in self.children)

    def __repr__(self):
        return '%s(%s, %s)' % (self.__class__.__name__, repr(self.rule), repr(self.children))


class NoSuchRuleException(AttributeError):
    """Rule not found (raised by :class:`AttrTree` for unknown children)."""

    def __init__(self, rule, tree: Optional[AttrTree] = None):
        post = '' if tree is None else f' ({repr(tree.rule)})'
        super().__init__(f'no {repr(rule)} child in tree{post}')


class GenericParser(ABC):
    """
    Generic parser using lark.

    Inherit to define a parser, say ThetaRecordParser for NONMEM, with the workflow:

    1. Lex and parse a 'buffer' using Lark_ (from :attr:`grammar` file). Builds AST.
    2. Convert to :class:`AttrTree` for immutability and convenience traversal.

    Attributes:
        self.buffer: Buffer parsed by :meth:`parse`.
        self.grammar: Path to grammar file.
        self.root: Root of final tree. Instance of :class:`AttrTree`.

    .. _Lark:
        https://github.com/lark-parser/lark
    """

    """:class:`AttrTree` implementation."""
    AttrTree = AttrTree

    lark: Lark
    lark_options = dict(
        start='root',
        parser='lalr',
        keep_all_tokens=True,
        propagate_positions=False,
        maybe_placeholders=False,
        debug=False,
        cache=False,
    )
    post_process: Tuple[Union[Visitor, Transformer, Callable[[str, Tree], Tree]], ...] = ()

    def __init__(self, buf=None):
        self.root = self.parse(buf)

    def parse(self, buf):
        """Parses a buffer, transforms and constructs :class:`AttrTree` object.

        Args:
            buf: Buffer to parse.
        """
        self.buffer = buf
        if self.buffer is None:
            return None

        root = self.lark.parse(self.buffer)

        for processor in self.post_process:
            if isinstance(processor, Visitor):
                processor.visit(root)
            elif isinstance(processor, Transformer):
                root = processor.transform(root)
            elif isinstance(processor, Callable):
                root = processor(self.buffer, root)
            else:
                raise TypeError(f'Processor {processor} must be a Visitor or a Transformer')

        return _from_lark_tree(root)

    def __str__(self):
        if not self.root:
            return repr(self)
        lines = str(prettyprint.transform(self.root)).splitlines()
        return '\n'.join(lines)


def _remove_token_and_space(node: Union[AttrTree, AttrToken], rule: str):
    if isinstance(node, AttrTree):
        return remove_token_and_space(node, rule, recursive=True)
    else:
        return node


def remove_token_and_space(tree: AttrTree, rule: str, recursive: bool = False):
    """Remove all tokens with rule and any WS before it"""
    new_nodes = []
    for node in tree.children:
        if node.rule == rule:
            if len(new_nodes) > 0 and new_nodes[-1].rule == 'WS':
                new_nodes.pop()
        else:
            new_nodes.append(node)

    if recursive:
        new_nodes = (_remove_token_and_space(node, rule) for node in new_nodes)

    return AttrTree(tree.rule, tuple(new_nodes))


def insert_before_or_at_end(
    tree: ImmutableTree, rule: str, nodes: Iterable[Union[ImmutableTree, ImmutableLeaf]]
):
    """Insert nodes before rule or if rule does not exist at end"""
    kept = []
    found = False
    for node in tree.children:
        if node.rule == rule:
            kept.extend(nodes)
            found = True
        kept.append(node)
    if not found:
        kept.extend(nodes)

    return AttrTree(tree.rule, tuple(kept))


def insert_after(
    tree: ImmutableTree, rule: str, nodes: Iterable[Union[ImmutableTree, ImmutableLeaf]]
):
    """Insert nodes after rule"""
    kept = []
    for node in tree.children:
        kept.append(node)
        if node.rule == rule:
            kept.extend(nodes)

    return AttrTree(tree.rule, tuple(kept))


def _from_lark_tree(tree: Tree) -> AttrTree:
    return AttrTree(
        tree.data,
        tuple(
            _from_lark_tree(child) if isinstance(child, Tree) else _from_lark_token(child)
            for child in tree.children
        ),
    )


def _from_lark_token(token: Token) -> AttrToken:
    return AttrToken(token.type, token.value)
